import asyncio
import time
from typing import Optional

from pylabrobot.machines.machine import Machine
from pylabrobot.resources import Coordinate, ResourceHolder

from .backend import TemperatureControllerBackend


class TemperatureController(ResourceHolder, Machine):
  """Temperature controller, for heating or for cooling."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    backend: TemperatureControllerBackend,
    child_location: Coordinate,
    category: str = "temperature_controller",
    model: Optional[str] = None,
  ):
    ResourceHolder.__init__(
      self,
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      child_location=child_location,
      category=category,
      model=model,
    )
    Machine.__init__(self, backend=backend)
    self.backend: TemperatureControllerBackend = backend  # fix type
    self.target_temperature: Optional[float] = None

  async def set_temperature(self, temperature: float, passive: bool = False):
    """Set the temperature of the temperature controller.

    Args:
      temperature: Temperature in Celsius.
      passive: If ``True`` and cooling is required, allow the device to cool
        down naturally without calling ``set_temperature`` on the backend.
        This can be used for backends that do not support active cooling or to
        explicitly disable active cooling when it is available.
    """
    current = await self.backend.get_current_temperature()

    self.target_temperature = temperature

    if temperature < current:
      if passive:  # if passive, we do nothing and return early.
        return

      # If we have to cool but the backend does not support active cooling,
      # and we are not passive cooling, raise an error.
      if not self.backend.supports_active_cooling:
        raise ValueError(
          "Backend does not support active cooling. Use passive=True to allow "
          "passive cooling or set a higher temperature."
        )

    return await self.backend.set_temperature(temperature)

  async def get_temperature(self) -> float:
    """Get the current temperature of the temperature controller in Celsius."""
    return await self.backend.get_current_temperature()

  async def wait_for_temperature(self, timeout: float = 300.0, tolerance: float = 0.5) -> None:
    """Wait for the temperature to reach the target temperature. The target temperature must be
    set by `set_temperature()`.

    Args:
      timeout: Timeout in seconds.
      tolerance: Tolerance in Celsius.
    """
    if self.target_temperature is None:
      raise RuntimeError("Target temperature is not set.")
    start = time.time()
    while time.time() - start < timeout:
      temperature = await self.get_temperature()
      if abs(temperature - self.target_temperature) < tolerance:
        return
      await asyncio.sleep(1.0)
    raise TimeoutError(f"Temperature did not reach target temperature within {timeout} seconds.")

  async def deactivate(self):
    """Deactivate the temperature controller. This will stop the heating or cooling, and return
    the temperature to ambient temperature. The target temperature will be reset to `None`.
    """
    self.target_temperature = None
    return await self.backend.deactivate()

  async def stop(self):
    """Stop the temperature controller and close the backend connection."""
    await self.deactivate()
    await super().stop()

  def serialize(self) -> dict:
    return {
      **Machine.serialize(self),
      **ResourceHolder.serialize(self),
    }



import warnings

from .molecular_devices.backend import (  # noqa: F401
  MolecularDevicesBackend,
  MolecularDevicesSettings,
)

warnings.warn(
  "pylabrobot.plate_reading.molecular_devices_backend is deprecated and will be removed in a future release. "
  "Please use pylabrobot.plate_reading.molecular_devices.molecular_devices_backend instead.",
)



from typing import List, Optional, Union

from pylabrobot.liquid_handling.backends.backend import (
  LiquidHandlerBackend,
)
from pylabrobot.liquid_handling.standard import (
  Drop,
  DropTipRack,
  MultiHeadAspirationContainer,
  MultiHeadAspirationPlate,
  MultiHeadDispenseContainer,
  MultiHeadDispensePlate,
  Pickup,
  PickupTipRack,
  ResourceDrop,
  ResourceMove,
  ResourcePickup,
  SingleChannelAspiration,
  SingleChannelDispense,
)
from pylabrobot.resources import Tip


class LiquidHandlerChatterboxBackend(LiquidHandlerBackend):
  """Chatter box backend for device-free testing. Prints out all operations."""

  _pip_length = 5
  _vol_length = 8
  _resource_length = 20
  _offset_length = 16
  _flow_rate_length = 10
  _blowout_length = 10
  _lld_z_length = 10
  _kwargs_length = 15
  _tip_type_length = 12
  _max_volume_length = 16
  _fitting_depth_length = 20
  _tip_length_length = 16
  # _pickup_method_length = 20
  _filter_length = 10

  def __init__(self, num_channels: int = 8):
    """Initialize a chatter box backend."""
    super().__init__()
    self._num_channels = num_channels
    self._num_arms = 1
    self._head96_installed = True

  async def setup(self):
    await super().setup()
    print("Setting up the liquid handler.")

  async def stop(self):
    print("Stopping the liquid handler.")

  def serialize(self) -> dict:
    return {**super().serialize(), "num_channels": self.num_channels}

  @property
  def num_channels(self) -> int:
    return self._num_channels

  async def pick_up_tips(self, ops: List[Pickup], use_channels: List[int], **backend_kwargs):
    print("Picking up tips:")
    header = (
      f"{'pip#':<{LiquidHandlerChatterboxBackend._pip_length}} "
      f"{'resource':<{LiquidHandlerChatterboxBackend._resource_length}} "
      f"{'offset':<{LiquidHandlerChatterboxBackend._offset_length}} "
      f"{'tip type':<{LiquidHandlerChatterboxBackend._tip_type_length}} "
      f"{'max volume (µL)':<{LiquidHandlerChatterboxBackend._max_volume_length}} "
      f"{'fitting depth (mm)':<{LiquidHandlerChatterboxBackend._fitting_depth_length}} "
      f"{'tip length (mm)':<{LiquidHandlerChatterboxBackend._tip_length_length}} "
      # f"{'pickup method':<{ChatterboxBackend._pickup_method_length}} "
      f"{'filter':<{LiquidHandlerChatterboxBackend._filter_length}}"
    )
    print(header)

    for op, channel in zip(ops, use_channels):
      offset = f"{round(op.offset.x, 1)},{round(op.offset.y, 1)},{round(op.offset.z, 1)}"
      row = (
        f"  p{channel}: "
        f"{op.resource.name[-30:]:<{LiquidHandlerChatterboxBackend._resource_length}} "
        f"{offset:<{LiquidHandlerChatterboxBackend._offset_length}} "
        f"{op.tip.__class__.__name__:<{LiquidHandlerChatterboxBackend._tip_type_length}} "
        f"{op.tip.maximal_volume:<{LiquidHandlerChatterboxBackend._max_volume_length}} "
        f"{op.tip.fitting_depth:<{LiquidHandlerChatterboxBackend._fitting_depth_length}} "
        f"{op.tip.total_tip_length:<{LiquidHandlerChatterboxBackend._tip_length_length}} "
        # f"{str(op.tip.pickup_method)[-20:]:<{ChatterboxBackend._pickup_method_length}} "
        f"{'Yes' if op.tip.has_filter else 'No':<{LiquidHandlerChatterboxBackend._filter_length}}"
      )
      print(row)

  async def drop_tips(self, ops: List[Drop], use_channels: List[int], **backend_kwargs):
    print("Dropping tips:")
    header = (
      f"{'pip#':<{LiquidHandlerChatterboxBackend._pip_length}} "
      f"{'resource':<{LiquidHandlerChatterboxBackend._resource_length}} "
      f"{'offset':<{LiquidHandlerChatterboxBackend._offset_length}} "
      f"{'tip type':<{LiquidHandlerChatterboxBackend._tip_type_length}} "
      f"{'max volume (µL)':<{LiquidHandlerChatterboxBackend._max_volume_length}} "
      f"{'fitting depth (mm)':<{LiquidHandlerChatterboxBackend._fitting_depth_length}} "
      f"{'tip length (mm)':<{LiquidHandlerChatterboxBackend._tip_length_length}} "
      # f"{'pickup method':<{ChatterboxBackend._pickup_method_length}} "
      f"{'filter':<{LiquidHandlerChatterboxBackend._filter_length}}"
    )
    print(header)

    for op, channel in zip(ops, use_channels):
      offset = f"{round(op.offset.x, 1)},{round(op.offset.y, 1)},{round(op.offset.z, 1)}"
      row = (
        f"  p{channel}: "
        f"{op.resource.name[-30:]:<{LiquidHandlerChatterboxBackend._resource_length}} "
        f"{offset:<{LiquidHandlerChatterboxBackend._offset_length}} "
        f"{op.tip.__class__.__name__:<{LiquidHandlerChatterboxBackend._tip_type_length}} "
        f"{op.tip.maximal_volume:<{LiquidHandlerChatterboxBackend._max_volume_length}} "
        f"{op.tip.fitting_depth:<{LiquidHandlerChatterboxBackend._fitting_depth_length}} "
        f"{op.tip.total_tip_length:<{LiquidHandlerChatterboxBackend._tip_length_length}} "
        # f"{str(op.tip.pickup_method)[-20:]:<{ChatterboxBackend._pickup_method_length}} "
        f"{'Yes' if op.tip.has_filter else 'No':<{LiquidHandlerChatterboxBackend._filter_length}}"
      )
      print(row)

  async def aspirate(
    self,
    ops: List[SingleChannelAspiration],
    use_channels: List[int],
    **backend_kwargs,
  ):
    print("Aspirating:")
    header = (
      f"{'pip#':<{LiquidHandlerChatterboxBackend._pip_length}} "
      f"{'vol(ul)':<{LiquidHandlerChatterboxBackend._vol_length}} "
      f"{'resource':<{LiquidHandlerChatterboxBackend._resource_length}} "
      f"{'offset':<{LiquidHandlerChatterboxBackend._offset_length}} "
      f"{'flow rate':<{LiquidHandlerChatterboxBackend._flow_rate_length}} "
      f"{'blowout':<{LiquidHandlerChatterboxBackend._blowout_length}} "
      f"{'lld_z':<{LiquidHandlerChatterboxBackend._lld_z_length}}  "
    )
    for key in backend_kwargs:
      header += f"{key:<{LiquidHandlerChatterboxBackend._kwargs_length}} "[-16:]
    print(header)

    for o, p in zip(ops, use_channels):
      offset = f"{round(o.offset.x, 1)},{round(o.offset.y, 1)},{round(o.offset.z, 1)}"
      row = (
        f"  p{p}: "
        f"{o.volume:<{LiquidHandlerChatterboxBackend._vol_length}} "
        f"{o.resource.name[-20:]:<{LiquidHandlerChatterboxBackend._resource_length}} "
        f"{offset:<{LiquidHandlerChatterboxBackend._offset_length}} "
        f"{str(o.flow_rate):<{LiquidHandlerChatterboxBackend._flow_rate_length}} "
        f"{str(o.blow_out_air_volume):<{LiquidHandlerChatterboxBackend._blowout_length}} "
        f"{str(o.liquid_height):<{LiquidHandlerChatterboxBackend._lld_z_length}} "
      )
      for key, value in backend_kwargs.items():
        if isinstance(value, list) and all(isinstance(v, bool) for v in value):
          value = "".join("T" if v else "F" for v in value)
        if isinstance(value, list):
          value = "".join(map(str, value))
        row += f" {value:<15}"
      print(row)

  async def dispense(
    self,
    ops: List[SingleChannelDispense],
    use_channels: List[int],
    **backend_kwargs,
  ):
    print("Dispensing:")
    header = (
      f"{'pip#':<{LiquidHandlerChatterboxBackend._pip_length}} "
      f"{'vol(ul)':<{LiquidHandlerChatterboxBackend._vol_length}} "
      f"{'resource':<{LiquidHandlerChatterboxBackend._resource_length}} "
      f"{'offset':<{LiquidHandlerChatterboxBackend._offset_length}} "
      f"{'flow rate':<{LiquidHandlerChatterboxBackend._flow_rate_length}} "
      f"{'blowout':<{LiquidHandlerChatterboxBackend._blowout_length}} "
      f"{'lld_z':<{LiquidHandlerChatterboxBackend._lld_z_length}}  "
    )
    for key in backend_kwargs:
      header += f"{key:<{LiquidHandlerChatterboxBackend._kwargs_length}} "[-16:]
    print(header)

    for o, p in zip(ops, use_channels):
      offset = f"{round(o.offset.x, 1)},{round(o.offset.y, 1)},{round(o.offset.z, 1)}"
      row = (
        f"  p{p}: "
        f"{o.volume:<{LiquidHandlerChatterboxBackend._vol_length}} "
        f"{o.resource.name[-20:]:<{LiquidHandlerChatterboxBackend._resource_length}} "
        f"{offset:<{LiquidHandlerChatterboxBackend._offset_length}} "
        f"{str(o.flow_rate):<{LiquidHandlerChatterboxBackend._flow_rate_length}} "
        f"{str(o.blow_out_air_volume):<{LiquidHandlerChatterboxBackend._blowout_length}} "
        f"{str(o.liquid_height):<{LiquidHandlerChatterboxBackend._lld_z_length}} "
      )
      for key, value in backend_kwargs.items():
        if isinstance(value, list) and all(isinstance(v, bool) for v in value):
          value = "".join("T" if v else "F" for v in value)
        if isinstance(value, list):
          value = "".join(map(str, value))
        row += f" {value:<{LiquidHandlerChatterboxBackend._kwargs_length}}"
      print(row)

  async def pick_up_tips96(self, pickup: PickupTipRack, **backend_kwargs):
    print(f"Picking up tips from {pickup.resource.name}.")

  async def drop_tips96(self, drop: DropTipRack, **backend_kwargs):
    print(f"Dropping tips to {drop.resource.name}.")

  async def aspirate96(
    self, aspiration: Union[MultiHeadAspirationPlate, MultiHeadAspirationContainer]
  ):
    if isinstance(aspiration, MultiHeadAspirationPlate):
      resource = aspiration.wells[0].parent
    else:
      resource = aspiration.container
    print(f"Aspirating {aspiration.volume} from {resource}.")

  async def dispense96(self, dispense: Union[MultiHeadDispensePlate, MultiHeadDispenseContainer]):
    if isinstance(dispense, MultiHeadDispensePlate):
      resource = dispense.wells[0].parent
    else:
      resource = dispense.container
    print(f"Dispensing {dispense.volume} to {resource}.")

  async def pick_up_resource(self, pickup: ResourcePickup):
    print(f"Picking up resource: {pickup}")

  async def move_picked_up_resource(self, move: ResourceMove):
    print(f"Moving picked up resource: {move}")

  async def drop_resource(self, drop: ResourceDrop):
    print(f"Dropping resource: {drop}")

  async def request_tip_presence(self) -> List[Optional[bool]]:
    """Return tip presence based on the tip tracker state.

    Returns:
      A list of length `num_channels` where each element is `True` if a tip is mounted,
      `False` if not, or `None` if unknown.
    """
    return [self.head[ch].has_tip for ch in range(self.num_channels)]

  def can_pick_up_tip(self, channel_idx: int, tip: Tip) -> bool:
    return True



"""Unit test validating center-center coordinates of Hamilton
tip spots across all PLR-integrated Hamilton tip racks.

These tests instantiate Hamilton tip rack models, compute each
rack's representative H1 tip spot position relative to the tip rack origin
and verify that the reported center-center coordinates match
expected reference values.
"""

import unittest

from pylabrobot.resources.carrier import Coordinate
from pylabrobot.resources.hamilton import (
  hamilton_96_tiprack_10uL,
  hamilton_96_tiprack_10uL_filter,
  hamilton_96_tiprack_50uL,
  hamilton_96_tiprack_50uL_filter,
  hamilton_96_tiprack_50uL_NTR,
  hamilton_96_tiprack_300uL,
  hamilton_96_tiprack_300uL_filter,
  hamilton_96_tiprack_300uL_filter_slim,
  hamilton_96_tiprack_300uL_filter_ultrawide,
  hamilton_96_tiprack_1000uL,
  hamilton_96_tiprack_1000uL_filter,
  hamilton_96_tiprack_1000uL_filter_ultrawide,
  hamilton_96_tiprack_1000uL_filter_wide,
)
from pylabrobot.resources.tip_rack import TipRack


class HamiltonTipSpotTests(unittest.TestCase):
  def test_tipspot_h1_cc(self):
    """Tests for PLR-integrated Hamilton TipRacks' accurate TipSpot center-center coordinates."""

    def check_tip_spot_h1(tr: TipRack, expect: Coordinate):
      h1_loc = tr.get_item("H1").get_absolute_location("c", "c")
      assert h1_loc.x == expect.x and h1_loc.y == expect.y, f"{h1_loc} != {expect}"

    common_tip_rack_loc = Coordinate(x=11.7, y=9.8, z=-22.5)
    check_tip_spot_h1(hamilton_96_tiprack_10uL_filter("tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_10uL(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_50uL_filter(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_50uL(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_300uL_filter(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_300uL(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_300uL_filter_slim(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_300uL_filter_ultrawide(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_1000uL_filter(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_1000uL(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_1000uL_filter_wide(name="tr"), common_tip_rack_loc)
    check_tip_spot_h1(hamilton_96_tiprack_1000uL_filter_ultrawide(name="tr"), common_tip_rack_loc)

    ntr_loc = Coordinate(x=13.525, y=11.625, z=13.5)
    check_tip_spot_h1(hamilton_96_tiprack_50uL_NTR(name="tr"), ntr_loc)



import unittest

from pylabrobot.plate_reading.tecan.spark20m.spark_packet_parser import (
  SparkPacket,
  SparkParser,
  parse_single_spark_packet,
)


class TestSparkPacket(unittest.TestCase):
  def test_packet_too_short(self) -> None:
    with self.assertRaises(ValueError) as ctx:
      SparkPacket(b"\x81\x00\x00")  # Only 3 bytes, need at least 5
    self.assertIn("too short", str(ctx.exception))

  def test_packet_payload_truncated(self) -> None:
    # Header says payload is 10 bytes, but we only provide 5
    with self.assertRaises(ValueError) as ctx:
      SparkPacket(b"\x81\x00\x00\x0a\x01\x02\x03\x04\x05")
    self.assertIn("shorter than indicated", str(ctx.exception))

  def test_valid_resp_ready_packet(self) -> None:
    # Type 129 (RespReady), seq 0, payload len 0, checksum
    data = bytes([129, 0, 0, 0, 129])  # checksum = XOR of header = 129
    packet = SparkPacket(data)
    self.assertEqual(packet.type, "RespReady")
    self.assertEqual(packet.seq_num, 0)
    self.assertEqual(packet.payload_len, 0)

  def test_valid_resp_message_packet(self) -> None:
    # Type 133 (RespMessage), seq 1, payload with message number and text
    payload = b"\x00\x64format|arg1"  # number=100, format="format", args=["arg1"]
    header = bytes([133, 1, 0, len(payload)])
    checksum = 0
    for b in header + payload:
      checksum ^= b
    data = header + payload + bytes([checksum])
    packet = SparkPacket(data)
    self.assertEqual(packet.type, "RespMessage")
    self.assertEqual(packet.parsed_payload["number"], 100)
    self.assertEqual(packet.parsed_payload["format"], "format")
    self.assertEqual(packet.parsed_payload["args"], ["arg1"])


class TestSparkParser(unittest.TestCase):
  def test_parser_handles_invalid_packets(self) -> None:
    # Mix of valid and invalid packets
    valid_packet = bytes([129, 0, 0, 0, 129])
    invalid_packet = b"\x81\x00"  # Too short

    parser = SparkParser([valid_packet, invalid_packet])

    # Should have parsed the valid one
    self.assertEqual(len(parser.all_packets), 1)
    self.assertEqual(parser.all_packets[0].type, "RespReady")

  def test_parser_empty_input(self) -> None:
    parser = SparkParser([])
    self.assertEqual(len(parser.all_packets), 0)
    self.assertEqual(parser.process_all_sequences(), {})


class TestParseSingleSparkPacket(unittest.TestCase):
  def test_returns_error_dict_on_failure(self) -> None:
    result = parse_single_spark_packet(b"\x00")  # Too short
    self.assertIn("error", result)
    self.assertIn("hex_string", result)

  def test_returns_packet_dict_on_success(self) -> None:
    data = bytes([129, 5, 0, 0, 129 ^ 5])  # RespReady, seq 5
    result = parse_single_spark_packet(data)
    self.assertEqual(result["type"], "RespReady")
    self.assertEqual(result["seq_num"], 5)


if __name__ == "__main__":
  unittest.main()



import tempfile
import unittest
from pathlib import Path

from pylabrobot import load_config
from pylabrobot.config.config import Config
from pylabrobot.config.formats import ConfigLoader, ConfigSaver
from pylabrobot.config.formats.ini_config import IniLoader, IniSaver
from pylabrobot.config.formats.json_config import (
  JsonLoader,
  JsonSaver,
)
from pylabrobot.config.io.file import FileReader, FileWriter


class ConfigTests(unittest.TestCase):
  """Tests for pylabrobot.config"""

  def run_file_reader_writer_test(
    self,
    format_loader: ConfigLoader,
    format_saver: ConfigSaver,
    write_to: Path,
    should_be: Config,
  ):
    writer = FileWriter(format_saver=format_saver)
    writer.write(write_to, should_be)
    reader = FileReader(format_loader=format_loader)
    cfg = reader.read(write_to)
    assert cfg == should_be

  def test_file_reader_writer(self):
    tmp_path: Path = Path(tempfile.mkdtemp())
    fake_config = Config(
      logging=Config.Logging(
        log_dir=tmp_path / "logs",
      )
    )
    cases = (
      (IniLoader(), IniSaver(), "fake_config.ini"),
      (JsonLoader(), JsonSaver(), "fake_config.json"),
    )
    for rdr, wr, fp in cases:
      self.run_file_reader_writer_test(rdr, wr, tmp_path / fp, fake_config)

  def test_load_config_creates_default(self):
    cwd = Path.cwd()
    test_path = cwd / "test_config.ini"
    if test_path.exists():
      test_path.unlink()
    assert not test_path.exists()
    cfg = load_config("test_config", create_default=True, create_module_level=False)
    assert test_path.exists()
    assert cfg == Config()

    (cwd / "test_config.ini").unlink()



import math

from pylabrobot.serializer import SerializableMixin
from pylabrobot.utils.linalg import matrix_multiply_3x3


class Rotation(SerializableMixin):
  """Represents a 3D rotation."""

  def __init__(self, x: float = 0, y: float = 0, z: float = 0):
    self.x = x  # around x-axis, roll
    self.y = y  # around y-axis, pitch
    self.z = z  # around z-axis, yaw

  def get_rotation_matrix(self):
    # Create rotation matrices for each axis
    Rz = [
      [
        math.cos(math.radians(self.z)),
        -math.sin(math.radians(self.z)),
        0,
      ],
      [
        math.sin(math.radians(self.z)),
        math.cos(math.radians(self.z)),
        0,
      ],
      [0, 0, 1],
    ]
    Ry = [
      [
        math.cos(math.radians(self.y)),
        0,
        math.sin(math.radians(self.y)),
      ],
      [0, 1, 0],
      [
        -math.sin(math.radians(self.y)),
        0,
        math.cos(math.radians(self.y)),
      ],
    ]
    Rx = [
      [1, 0, 0],
      [
        0,
        math.cos(math.radians(self.x)),
        -math.sin(math.radians(self.x)),
      ],
      [
        0,
        math.sin(math.radians(self.x)),
        math.cos(math.radians(self.x)),
      ],
    ]
    # Combine rotations: The order of multiplication matters and defines the behavior significantly.
    # This is a common order: Rz * Ry * Rx
    return matrix_multiply_3x3(matrix_multiply_3x3(Rz, Ry), Rx)

  def __str__(self) -> str:
    return f"Rotation(x={self.x}, y={self.y}, z={self.z})"

  def __add__(self, other) -> "Rotation":
    return Rotation(x=self.x + other.x, y=self.y + other.y, z=self.z + other.z)

  @staticmethod
  def deserialize(data) -> "Rotation":
    return Rotation(data["x"], data["y"], data["z"])

  def __repr__(self) -> str:
    return self.__str__()

  @property
  def roll(self) -> float:
    return self.x

  @property
  def pitch(self) -> float:
    return self.y

  @property
  def yaw(self) -> float:
    return self.z



"""Defines audio that machines can generate.

Source of audio: https://simpleguics2pygame.readthedocs.io/en/latest/_static/links/snd_links.html
"""

import functools

try:
  from IPython.display import Audio, display

  USE_AUDIO = True
except ImportError as e:
  USE_AUDIO = False
  _AUDIO_IMPORT_ERROR = e


def _audio_check(func):
  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    if not USE_AUDIO:
      raise RuntimeError(
        f"Audio functionality requires IPython.display. Import error: {_AUDIO_IMPORT_ERROR}"
      )
    return func(*args, **kwargs)

  return wrapper


# ====== 1. Identifying items on deck (e.g. through LLD or Z-drive engagement) ======


@_audio_check
def play_not_found():
  display(
    Audio(
      url="https://codeskulptor-demos.commondatastorage.googleapis.com/pang/arrow.mp3",
      autoplay=True,
    )
  )


@_audio_check
def play_got_item():
  display(
    Audio(
      url="https://codeskulptor-demos.commondatastorage.googleapis.com/descent/gotitem.mp3",
      autoplay=True,
    )
  )



"""EL406 exception classes.

This module contains exception classes used by the BioTek EL406
plate washer backend.
"""

from __future__ import annotations


class EL406CommunicationError(Exception):
  """Exception raised for FTDI/USB communication errors with the EL406.

  This exception is raised when low-level communication fails, such as:
  - USB device disconnected
  - FTDI driver errors
  - Write/read failures

  Attributes:
    operation: The operation that failed (e.g., "write", "read", "open").
    original_error: The underlying exception that caused this error.
  """

  def __init__(
    self,
    message: str,
    operation: str = "",
    original_error: Exception | None = None,
  ) -> None:
    super().__init__(message)
    self.operation = operation
    self.original_error = original_error


class EL406DeviceError(Exception):
  """Exception raised when the EL406 device reports an error via the validity field.

  The device returns a non-zero validity code in the status poll response
  when a step command fails (e.g., no buffer fluid, invalid syringe, hardware fault).

  Attributes:
    error_code: The raw error code from the device (e.g., 0x1500).
    message: Human-readable error description.
  """

  def __init__(self, error_code: int, message: str) -> None:
    self.error_code = error_code
    super().__init__(f"EL406 error 0x{error_code:04X}: {message}")



import sys
import unittest
from typing import List

from pylabrobot.resources import (
  Coordinate,
  ItemizedResource,
  Plate,
  Resource,
  Well,
  create_equally_spaced_2d,
  create_ordered_items_2d,
)
from pylabrobot.resources.plate import Lid

if sys.version_info >= (3, 8):
  from typing import Literal
else:
  from typing_extensions import Literal


class TestItemizedResource(unittest.TestCase):
  def setUp(self) -> None:
    self.plate = Plate(
      "plate",
      size_x=1,
      size_y=1,
      size_z=1,
      ordered_items=create_ordered_items_2d(
        Well,
        num_items_x=12,
        num_items_y=8,
        dx=0,
        dy=0,
        dz=0,
        item_dx=9,
        item_dy=9,
        size_x=9,
        size_y=9,
        size_z=9,
      ),
    )
    self.plate.location = Coordinate.zero()
    return super().setUp()

  def test_initialize_with_wells(self):
    self.assertEqual(len(self.plate.children), 96)
    self.assertEqual(self.plate.children[0].name, "plate_well_A1")
    self.assertEqual(self.plate.children[95].name, "plate_well_H12")

  def test_get_item_int(self):
    self.assertEqual(self.plate.get_item(0).name, "plate_well_A1")
    self.assertEqual(self.plate.get_item(95).name, "plate_well_H12")

  def test_get_item_str(self):
    self.assertEqual(self.plate.get_item("A1").name, "plate_well_A1")
    self.assertEqual(self.plate.get_item("B1").name, "plate_well_B1")
    self.assertEqual(self.plate.get_item("A2").name, "plate_well_A2")

  def test_get_item_tuple(self):
    self.assertEqual(self.plate.get_item((0, 0)).name, "plate_well_A1")
    self.assertEqual(self.plate.get_item((7, 11)).name, "plate_well_H12")

  def test_well_get_absolute_location(self):
    self.assertEqual(
      self.plate.get_item(0).get_absolute_location(),
      Coordinate(0, 63, 0),
    )
    self.assertEqual(
      self.plate.get_item(7).get_absolute_location(),
      Coordinate(0, 0, 0),
    )
    self.assertEqual(
      self.plate.get_item(8).get_absolute_location(),
      Coordinate(9, 63, 0),
    )
    self.assertEqual(
      self.plate.get_item(17).get_absolute_location(),
      Coordinate(18, 54, 0),
    )
    self.assertEqual(
      self.plate.get_item(95).get_absolute_location(),
      Coordinate(99, 0, 0),
    )

  def test_getitem_int(self):
    self.assertEqual(self.plate[0][0].name, "plate_well_A1")

  def test_getitem_str(self):
    self.assertEqual(self.plate["A1"][0].name, "plate_well_A1")
    self.assertEqual(self.plate["B2"][0].name, "plate_well_B2")

  def test_getitem_slice(self):
    self.assertEqual(
      [w.name for w in self.plate[0:7]],
      [
        "plate_well_A1",
        "plate_well_B1",
        "plate_well_C1",
        "plate_well_D1",
        "plate_well_E1",
        "plate_well_F1",
        "plate_well_G1",
      ],
    )

  def test_getitem_range(self):
    self.assertEqual(
      [w.name for w in self.plate[range(7)]],
      [
        "plate_well_A1",
        "plate_well_B1",
        "plate_well_C1",
        "plate_well_D1",
        "plate_well_E1",
        "plate_well_F1",
        "plate_well_G1",
      ],
    )

  def test_getitem_str_range(self):
    self.assertEqual(
      [w.name for w in self.plate["A1:B2"]],
      [
        "plate_well_A1",
        "plate_well_A2",
        "plate_well_B1",
        "plate_well_B2",
      ],
    )

  def test_getitem_str_error(self):
    with self.assertRaises(IndexError):
      _ = self.plate["A13"]
    with self.assertRaises(IndexError):
      _ = self.plate["T1"]

  def test_getitem_tuple_int(self):
    self.assertEqual(
      [w.name for w in self.plate[0, 4, 1]],
      ["plate_well_A1", "plate_well_E1", "plate_well_B1"],
    )

  def test_getitem_tuple_str(self):
    self.assertEqual(
      [w.name for w in self.plate["A1", "B2", "A2"]],
      ["plate_well_A1", "plate_well_B2", "plate_well_A2"],
    )

  def test_get_row(self):
    self.assertEqual(
      [w.name for w in self.plate.row(0)],
      [
        "plate_well_A1",
        "plate_well_A2",
        "plate_well_A3",
        "plate_well_A4",
        "plate_well_A5",
        "plate_well_A6",
        "plate_well_A7",
        "plate_well_A8",
        "plate_well_A9",
        "plate_well_A10",
        "plate_well_A11",
        "plate_well_A12",
      ],
    )

    self.assertEqual(
      [w.name for w in self.plate.row(3)],
      [
        "plate_well_D1",
        "plate_well_D2",
        "plate_well_D3",
        "plate_well_D4",
        "plate_well_D5",
        "plate_well_D6",
        "plate_well_D7",
        "plate_well_D8",
        "plate_well_D9",
        "plate_well_D10",
        "plate_well_D11",
        "plate_well_D12",
      ],
    )

  def test_get_row_str(self):
    self.assertEqual(
      [w.name for w in self.plate.row("A")],
      [w.name for w in self.plate.row(0)],
    )
    self.assertEqual(
      [w.name for w in self.plate.row("d")],
      [w.name for w in self.plate.row(3)],
    )

  def test_get_row_str_invalid(self):
    with self.assertRaises(ValueError):
      self.plate.row("Q")
    with self.assertRaises(ValueError):
      self.plate.row("AA")

  def test_get_column(self):
    self.assertEqual(
      [w.name for w in self.plate.column(0)],
      [
        "plate_well_A1",
        "plate_well_B1",
        "plate_well_C1",
        "plate_well_D1",
        "plate_well_E1",
        "plate_well_F1",
        "plate_well_G1",
        "plate_well_H1",
      ],
    )

    self.assertEqual(
      [w.name for w in self.plate.column(3)],
      [
        "plate_well_A4",
        "plate_well_B4",
        "plate_well_C4",
        "plate_well_D4",
        "plate_well_E4",
        "plate_well_F4",
        "plate_well_G4",
        "plate_well_H4",
      ],
    )


class TestItemizedResource1536(unittest.TestCase):
  """Test ItemizedResource with a 1536-well plate (>26 rows)."""

  plate: Plate

  @classmethod
  def setUpClass(cls) -> None:
    cls.plate = Plate(
      "plate_1536",
      size_x=127.8,
      size_y=85.5,
      size_z=8.6,
      ordered_items=create_ordered_items_2d(
        Well,
        num_items_x=48,
        num_items_y=32,
        dx=9.85,
        dy=6.75,
        dz=0.0,
        item_dx=2.3,
        item_dy=2.3,
        size_x=2.3,
        size_y=2.3,
        size_z=8.6,
      ),
    )

  def test_num_items(self):
    self.assertEqual(self.plate.num_items, 1536)
    self.assertEqual(self.plate.num_items_x, 48)
    self.assertEqual(self.plate.num_items_y, 32)

  def test_identifiers(self):
    keys = list(self.plate._ordering.keys())
    self.assertEqual(keys[0], "A1")
    self.assertEqual(keys[25], "Z1")
    self.assertEqual(keys[26], "AA1")
    self.assertEqual(keys[31], "AF1")
    self.assertEqual(keys[-1], "AF48")

  def test_get_item_str(self):
    self.assertEqual(self.plate.get_item("A1").name, "plate_1536_well_A1")
    self.assertEqual(self.plate.get_item("AA1").name, "plate_1536_well_AA1")
    self.assertEqual(self.plate.get_item("AF48").name, "plate_1536_well_AF48")

  def test_get_item_tuple(self):
    self.assertEqual(self.plate.get_item((0, 0)).name, "plate_1536_well_A1")
    self.assertEqual(self.plate.get_item((26, 0)).name, "plate_1536_well_AA1")
    self.assertEqual(self.plate.get_item((31, 47)).name, "plate_1536_well_AF48")

  def test_row_by_str(self):
    row_aa = self.plate.row("AA")
    self.assertEqual(len(row_aa), 48)
    self.assertEqual(row_aa[0].name, "plate_1536_well_AA1")
    self.assertEqual(row_aa[47].name, "plate_1536_well_AA48")

  def test_row_by_int(self):
    row_26 = self.plate.row(26)
    self.assertEqual(row_26[0].name, "plate_1536_well_AA1")

  def test_row_out_of_range(self):
    with self.assertRaises(ValueError):
      self.plate.row("AG")
    with self.assertRaises(ValueError):
      self.plate.row(32)

  def test_get_child_row(self):
    well_aa1 = self.plate.get_item("AA1")
    self.assertEqual(self.plate.get_child_row(well_aa1), 26)

    well_af48 = self.plate.get_item("AF48")
    self.assertEqual(self.plate.get_child_row(well_af48), 31)


class TestItemizedResourceWithLid(unittest.TestCase):
  def setUp(self) -> None:
    self.plate = Plate(
      "plate",
      size_x=127,
      size_y=86,
      size_z=15,
      lid=Lid("lid", size_x=127, size_y=86, size_z=10, nesting_z_height=2),
      ordered_items=create_ordered_items_2d(
        Well,
        num_items_x=12,
        num_items_y=8,
        dx=0,
        dy=0,
        dz=0,
        item_dx=9,
        item_dy=9,
        size_x=9,
        size_y=9,
        size_z=9,
      ),
    )

  def test_num_items_excludes_lid(self):
    self.assertEqual(self.plate.num_items, 96)

  def test_get_all_items_excludes_lid(self):
    items = self.plate.get_all_items()
    self.assertEqual(len(items), 96)
    for item in items:
      self.assertNotIsInstance(item, Lid)


class TestCreateEquallySpaced(unittest.TestCase):
  """Test for create_ordered_items_2d function."""

  def test_create_equally_spaced(self):
    self.maxDiff = None
    equally_spaced = create_equally_spaced_2d(
      Well,
      num_items_x=3,
      num_items_y=2,
      dx=0,
      dy=0,
      dz=0,
      item_dx=9,
      item_dy=9,
      size_x=9,
      size_y=9,
      size_z=9,
    )

    # assert that ids of items are correct
    ids = [id(item) for item in equally_spaced]
    self.assertEqual(len(ids), len(set(ids)))

    self.assertEqual(len(equally_spaced), 3)
    self.assertEqual(len(equally_spaced[0]), 2)
    self.assertEqual(len(equally_spaced[1]), 2)
    self.assertEqual(len(equally_spaced[2]), 2)

    correct_items = [
      [
        Well("well_0_0", size_x=9, size_y=9, size_z=9),
        Well("well_0_1", size_x=9, size_y=9, size_z=9),
      ],
      [
        Well("well_1_0", size_x=9, size_y=9, size_z=9),
        Well("well_1_1", size_x=9, size_y=9, size_z=9),
      ],
      [
        Well("well_2_0", size_x=9, size_y=9, size_z=9),
        Well("well_2_1", size_x=9, size_y=9, size_z=9),
      ],
    ]
    correct_items[0][0].location = Coordinate(0, 9, 0)
    correct_items[0][1].location = Coordinate(0, 0, 0)
    correct_items[1][0].location = Coordinate(9, 9, 0)
    correct_items[1][1].location = Coordinate(9, 0, 0)
    correct_items[2][0].location = Coordinate(18, 9, 0)
    correct_items[2][1].location = Coordinate(18, 0, 0)

    self.assertEqual(equally_spaced, correct_items)


class TestItemizedResourceTraversal(unittest.TestCase):
  def setUp(self) -> None:
    def item_id(row, column):
      return "ABC"[row] + str(column + 1)

    items = [
      Resource(name=item_id(row, col), size_x=1, size_y=1, size_z=1).at(Coordinate(col, row, 0))
      for col in range(3)
      for row in range(3)
    ]
    self.ir = ItemizedResource(
      "test_resource",
      ordered_items={item.name: item for item in items},
      size_x=3,
      size_y=3,
      size_z=1,
    )

  def _traverse_test(
    self,
    direction: Literal[
      "up",
      "down",
      "right",
      "left",
      "snake_up",
      "snake_down",
      "snake_left",
      "snake_right",
    ],
    start: Literal["top_left", "bottom_left", "top_right", "bottom_right"],
    pattern: List[str],
  ):
    items = []
    for wells in self.ir.traverse(batch_size=1, direction=direction, start=start, repeat=False):
      items.extend(wells)

    self.assertEqual(len(items), len(pattern))
    for w, identifier in zip(items, pattern):
      assert w.parent is not None
      self.assertEqual(w.name, w.parent.name + "_" + identifier)

  def test_traverse_down_top_left(self):
    self._traverse_test(
      direction="down",
      start="top_left",
      pattern=["A1", "B1", "C1", "A2", "B2", "C2", "A3", "B3", "C3"],
    )

  def test_traverse_down_top_right(self):
    self._traverse_test(
      direction="down",
      start="top_right",
      pattern=["A3", "B3", "C3", "A2", "B2", "C2", "A1", "B1", "C1"],
    )

  def test_traverse_up_bottom_left(self):
    self._traverse_test(
      direction="up",
      start="bottom_left",
      pattern=["C1", "B1", "A1", "C2", "B2", "A2", "C3", "B3", "A3"],
    )

  def test_traverse_up_bottom_right(self):
    self._traverse_test(
      direction="up",
      start="bottom_right",
      pattern=["C3", "B3", "A3", "C2", "B2", "A2", "C1", "B1", "A1"],
    )

  def test_traverse_right_top_left(self):
    self._traverse_test(
      direction="right",
      start="top_left",
      pattern=["A1", "A2", "A3", "B1", "B2", "B3", "C1", "C2", "C3"],
    )

  def test_traverse_right_bottom_left(self):
    self._traverse_test(
      direction="right",
      start="bottom_left",
      pattern=["C1", "C2", "C3", "B1", "B2", "B3", "A1", "A2", "A3"],
    )

  def test_traverse_left_top_right(self):
    self._traverse_test(
      direction="left",
      start="top_right",
      pattern=["A3", "A2", "A1", "B3", "B2", "B1", "C3", "C2", "C1"],
    )

  def test_traverse_left_bottom_right(self):
    self._traverse_test(
      direction="left",
      start="bottom_right",
      pattern=["C3", "C2", "C1", "B3", "B2", "B1", "A3", "A2", "A1"],
    )

  def test_traverse_snake_down_top_left(self):
    self._traverse_test(
      direction="snake_down",
      start="top_left",
      pattern=["A1", "B1", "C1", "C2", "B2", "A2", "A3", "B3", "C3"],
    )

  def test_traverse_snake_down_top_right(self):
    self._traverse_test(
      direction="snake_down",
      start="top_right",
      pattern=["A3", "B3", "C3", "C2", "B2", "A2", "A1", "B1", "C1"],
    )

  def test_traverse_snake_up_bottom_left(self):
    self._traverse_test(
      direction="snake_up",
      start="bottom_left",
      pattern=["C1", "B1", "A1", "A2", "B2", "C2", "C3", "B3", "A3"],
    )

  def test_traverse_snake_up_bottom_right(self):
    self._traverse_test(
      direction="snake_up",
      start="bottom_right",
      pattern=["C3", "B3", "A3", "A2", "B2", "C2", "C1", "B1", "A1"],
    )

  def test_traverse_snake_right_top_left(self):
    self._traverse_test(
      direction="snake_right",
      start="top_left",
      pattern=["A1", "A2", "A3", "B3", "B2", "B1", "C1", "C2", "C3"],
    )

  def test_traverse_snake_right_bottom_left(self):
    self._traverse_test(
      direction="snake_right",
      start="bottom_left",
      pattern=["C1", "C2", "C3", "B3", "B2", "B1", "A1", "A2", "A3"],
    )

  def test_traverse_snake_left_top_right(self):
    self._traverse_test(
      direction="snake_left",
      start="top_right",
      pattern=["A3", "A2", "A1", "B1", "B2", "B3", "C3", "C2", "C1"],
    )

  def test_traverse_snake_left_bottom_right(self):
    self._traverse_test(
      direction="snake_left",
      start="bottom_right",
      pattern=["C3", "C2", "C1", "B1", "B2", "B3", "A3", "A2", "A1"],
    )

  def test_traverse_down_repeat(self):
    # do a down traversal with batch size 5, which does not divide 9. `ItemizedResource.traverse`
    # should continue the batch from the beginning of the resource.
    # Go over the resource 10 times in total.

    items = []
    num_rounds = 10
    total_num_batches = 9 * num_rounds // 5  # 10 rounds, batch size 5
    num_batches = 0
    for batch in self.ir.traverse(batch_size=5, direction="down", start="top_left", repeat=True):
      if num_batches == total_num_batches:
        break
      items.extend(batch)
      self.assertEqual(len(batch), 5)
      num_batches += 1
    self.assertEqual(len(items), self.ir.num_items * num_rounds)



from pylabrobot.centrifuge.backend import CentrifugeBackend, LoaderBackend


class CentrifugeChatterboxBackend(CentrifugeBackend):
  async def setup(self):
    print("Setting up")

  async def stop(self):
    print("Stopping")

  async def open_door(self):
    print("Opening door")

  async def close_door(self):
    print("Closing door")

  async def lock_door(self):
    print("Locking door")

  async def unlock_door(self):
    print("Unlocking door")

  async def go_to_bucket1(self):
    print("Going to bucket 1")

  async def go_to_bucket2(self):
    print("Going to bucket 2")

  async def rotate_distance(self, distance):
    print(f"Rotating distance: {distance}")

  async def lock_bucket(self):
    print("Locking bucket")

  async def unlock_bucket(self):
    print("Unlocking bucket")

  async def spin(self, g: float, duration: float, acceleration: float):
    print(f"Starting spin cycle with g: {g}, duration: {duration}, acceleration: {acceleration}")


class LoaderChatterboxBackend(LoaderBackend):
  async def setup(self):
    print("Setting up")

  async def stop(self):
    print("Stopping")

  async def load(self):
    print("Loading")

  async def unload(self):
    print("Unloading")



"""Resource definitions for Eppendorf tubes."""

import warnings

from pylabrobot.resources.tube import Tube, TubeBottomType

# --------------------------------------------------------------------------- #
# 1.5 mL Eppendorf Tubes
# --------------------------------------------------------------------------- #


def Eppendorf_DNA_LoBind_1_5ml_Vb(name: str, model="Eppendorf_DNA_LoBind_1_5ml_Vb") -> Tube:
  """1.5 mL round-bottom snap-cap Eppendorf tube.

  .. deprecated:: Will be removed in 2026-04
      Use :func:`eppendorf_tube_1500uL_Vb` instead.

  cat. no.: 022431021 (Eppendorf™ DNA LoBind™ Tubes)

  - bottom_type=TubeBottomType.V
  - snap-cap lid
  """
  warnings.warn(
    "Eppendorf_DNA_LoBind_1_5ml_Vb is deprecated and will be removed in 2026-04. "
    "Use eppendorf_tube_1500uL_Vb instead.",
    DeprecationWarning,
    stacklevel=2,
  )
  return eppendorf_tube_1500uL_Vb(name=name)


def Eppendorf_Protein_LoBind_1_5ml_Vb(name: str) -> Tube:
  """1.5 mL round-bottom screw-cap Eppendorf tube.

  .. deprecated:: Will be removed in 2026-04
      Use :func:`eppendorf_tube_1500uL_Vb` instead.

  cat. no.: 022431081 (Eppendorf™ Protein LoBind™ Tubes)

  Same as Eppendorf_DNA_LoBind_1_5ml_Vb
  """
  warnings.warn(
    "Eppendorf_Protein_LoBind_1_5ml_Vb is deprecated and will be removed in 2026-04. "
    "Use eppendorf_tube_1500uL_Vb instead.",
    DeprecationWarning,
    stacklevel=2,
  )
  return eppendorf_tube_1500uL_Vb(name=name)


def eppendorf_tube_1500uL_Vb(name: str) -> Tube:
  """Eppendorf cat. no.: 022363204

  1.5 mL or 1_500 uL snap-cap Eppendorf tube with V-bottom;
  nickname: the original "Eppi".

  - Colour: transparent
  - alternative cat. no.:
    - 022431021: DNA LoBind®
    - 022431081: Protein LoBind®
  - Material: Polypropylene
  - Sterilized: ?
  - Autoclavable: Yes ("when open (121 °C, 20 min)")
  - Chemical resistance:?
  - Thermal resistance: ?
  - Sealing options: na
  - Centrifugation safety: Maximum safety and stability for centrifugation up to 25,000 x g
  - Total volume = 1_500 ul
  - URL: https://www.eppendorf.com/us-en/Products/Lab-Consumables/Lab-Tubes/Eppibr-Eppendorf-Safe-Lock-Tubes-p-022363204
  - technical drawing: ./engineering_diagrams/Eppendorf_1500uL_snapcap_tube.pdf
  """
  diameter = 9.85  # measured
  return Tube(
    name=name,
    size_x=diameter,
    size_y=diameter,
    size_z=38.9,  # measured
    model=eppendorf_tube_1500uL_Vb.__name__,
    max_volume=1_500,  # units: ul
    material_z_thickness=1.3,
    bottom_type=TubeBottomType.V,
    # compute_volume_from_height=_compute_volume_from_height_eppendorf_tube_1500uL_Vb, TODO
    # compute_height_from_volume=_compute_height_from_volume_eppendorf_tube_1500uL_Vb, TODO
  )


# --------------------------------------------------------------------------- #
# 2 mL Eppendorf Tubes
# --------------------------------------------------------------------------- #


def Eppendorf_DNA_LoBind_2ml_Ub(name: str) -> Tube:
  """2 mL round-bottom snap-cap Eppendorf tube. cat. no.: 022431048

  - bottom_type=TubeBottomType.U
  - snap-cap lid
  """
  diameter = 10.33  # measured
  return Tube(
    name=name,
    size_x=diameter,
    size_y=diameter,
    size_z=41,  # measured
    model="Eppendorf_DNA_LoBind_2ml_Ub",
    max_volume=2000,  # units: ul
    material_z_thickness=0.8,  # measured
  )


# --------------------------------------------------------------------------- #
# 5 mL Eppendorf Tubes
# --------------------------------------------------------------------------- #

# Calibration data: height (mm) → volume (µL).
# Obtained via ztouch probing of cavity_bottom, manual addition of known volumes,
# and LLD measurement of liquid height relative to cavity_bottom.
_eppendorf_tube_5mL_Vb_snapcap_height_volume_data = {
  0.0: 0.0,
  1.342: 5.0,
  1.875: 10.0,
  4.975: 50.0,
  6.975: 100.0,
  9.909: 200.0,
  13.742: 400.0,
  16.242: 600.0,
  18.375: 800.0,
  20.142: 1000.0,
  21.809: 1200.0,
  23.842: 1500.0,
  27.242: 2000.0,
  34.242: 3000.0,
  41.009: 4000.0,
  44.175: 4500.0,
  47.509: 5000.0,
  50.675: 5500.0,
}


def eppendorf_tube_5mL_Vb_snapcap(name: str) -> Tube:
  """Eppendorf cat. no.: 0030119401

  5 mL snap-cap Eppendorf tube with V-bottom.

  - Colour: transparent
  - Material: Polypropylene
  - Sterilized: No
  - Autoclavable: No
  - Chemical resistance:?
  - Thermal resistance: ?
  - Surface treatment: available as LoBind™ version
  - Sealing options: na
  - Centrifugation safety: Maximum safety and stability for centrifugation up to 25,000 x g
  - Total volume = 5_000 ul
  - URL: https://www.eppendorf.com/se-en/Products/Lab-Consumables/Lab-Tubes/EppendorfTubes-50mL-p-0030119401
  - technical drawing: ./engineering_diagrams/Eppendorf_5mL_snapcap_tube.pdf
  """
  diameter = 16.7
  return Tube(
    name=name,
    size_x=diameter,
    size_y=diameter,
    size_z=56.7,
    model=eppendorf_tube_5mL_Vb_snapcap.__name__,
    max_volume=5_000,  # units: ul
    material_z_thickness=1.2,
    bottom_type=TubeBottomType.V,
    height_volume_data=_eppendorf_tube_5mL_Vb_snapcap_height_volume_data,
  )



from pylabrobot.resources.carrier import PlateHolder
from pylabrobot.resources.plate import Plate
from pylabrobot.storage.backend import IncubatorBackend


class IncubatorChatterboxBackend(IncubatorBackend):
  def __init__(self):
    self._dummy_temperature = 37.0

  async def setup(self):
    print("Setting up incubator backend")

  async def stop(self):
    print("Stopping incubator backend")

  async def open_door(self):
    print("Opening door")

  async def close_door(self):
    print("Closing door")

  async def fetch_plate_to_loading_tray(self, plate: Plate, **backend_kwargs):
    print(f"Fetching plate {plate} to loading tray")

  async def take_in_plate(self, plate: Plate, site: PlateHolder, **backend_kwargs):
    print(f"Taking in plate {plate} at site {site}")

  async def set_temperature(self, temperature: float):
    print(f"Setting temperature to {temperature}")

  async def get_temperature(self) -> float:
    print("Getting temperature")
    return self._dummy_temperature

  async def start_shaking(self, frequency: float):
    print(f"Starting shaking at {frequency} Hz")

  async def stop_shaking(self):
    print("Stopping shaking")



import unittest
from typing import List
from unittest.mock import AsyncMock

from pylabrobot.powder_dispensing.backend import (
  DispenseResults,
  PowderDispense,
  PowderDispenserBackend,
)
from pylabrobot.powder_dispensing.powder_dispenser import (
  PowderDispenser,
)
from pylabrobot.resources import Cor_96_wellplate_360ul_Fb, Powder


class MockPowderDispenserBackend(PowderDispenserBackend):
  """A mock backend for testing."""

  async def setup(self) -> None:
    pass

  async def stop(self) -> None:
    pass

  async def dispense(
    self,
    dispense_parameters: List[PowderDispense],
    **backend_kwargs: None,
  ) -> List[DispenseResults]:
    results = []

    for dp in dispense_parameters:
      results.append(DispenseResults(actual_amount=dp.amount))

    return results


class TestPowderDispenser(unittest.IsolatedAsyncioTestCase):
  """
  Test class for PowderDispenser.
  """

  async def asyncSetUp(self) -> None:
    self.backend = AsyncMock(spec=MockPowderDispenserBackend)
    self.dispenser = PowderDispenser(backend=self.backend)
    await self.dispenser.setup()

  async def test_dispense_single_resource(self):
    plate = Cor_96_wellplate_360ul_Fb(name="test_resource")
    powder = Powder("salt")
    await self.dispenser.dispense(plate["A1"], powder, 0.005)
    self.backend.dispense.assert_called_once()

  async def test_dispense_multiple_resources(self):
    plate = Cor_96_wellplate_360ul_Fb(name="test_resource")
    resources = plate["A1"] + plate["A2"]
    powders = [Powder("salt"), Powder("salt")]
    amounts = [0.005, 0.010]
    await self.dispenser.dispense(resources, powders, amounts)
    self.assertEqual(self.backend.dispense.call_count, 1)

  async def test_dispense_parameters_handling(self):
    plate = Cor_96_wellplate_360ul_Fb(name="test_resource")
    powder = Powder("salt")
    dispense_parameters = {"param1": "value1"}
    await self.dispenser.dispense(
      plate["A1"],
      powder,
      0.005,
      dispense_parameters=dispense_parameters,
    )
    self.backend.dispense.assert_called_once()

  async def test_assertion_for_mismatched_lengths(self):
    with self.assertRaises(AssertionError):
      plate = Cor_96_wellplate_360ul_Fb(name="test_resource")
      list_of_powders = [Powder("salt"), Powder("salt")]
      await self.dispenser.dispense(plate["A1"], list_of_powders, [0.005])

    with self.assertRaises(AssertionError):
      plate = Cor_96_wellplate_360ul_Fb(name="test_resource")
      await self.dispenser.dispense(
        plate["A1"],
        Powder("salt"),
        [0.005, 0.010],
        dispense_parameters=[{"param": "value"}, {"param": "value"}],
      )


if __name__ == "__main__":
  unittest.main()



from pylabrobot.resources.trough import Trough


def ThermoFisherMatrixTrough8094(name: str) -> Trough:
  """Thermo Fisher Trough 8094 - 25mL. Part number 8094.
  https://www.thermofisher.com/order/catalog/product/8094
  """
  return Trough(
    name=name,
    size_x=147,
    size_y=58,
    size_z=27,
    max_volume=25000,
    model="Thermo Fisher 8094",
  )



from typing import Optional

from pylabrobot.io.capture import CaptureReader, capturer
from pylabrobot.io.ftdi import FTDI, FTDIValidator
from pylabrobot.io.hid import HID, HIDValidator
from pylabrobot.io.serial import Serial, SerialValidator
from pylabrobot.io.usb import USB, USBValidator
from pylabrobot.machines.backend import MachineBackend

cr: Optional[CaptureReader] = None


def validate(capture_file: str):
  """Start validation against a capture file.

  Args:
    capture_file: path to the capture file. Generate with start_capture.
  """

  if capturer.capture_active:
    raise RuntimeError("Cannot validate while capture is active")

  global cr
  cr = CaptureReader(path=capture_file)

  def _replace_io(obj):
    io2v = {
      USB: USBValidator,
      Serial: SerialValidator,
      FTDI: FTDIValidator,
      HID: HIDValidator,
    }
    if not hasattr(obj, "io"):
      return False
    if obj.io.__class__ in io2v:
      obj.io = io2v[obj.io.__class__](**obj.io.serialize(), cr=cr)
    elif obj.io.__class__ in io2v.values():
      obj.io.cr = cr
    else:
      return False
    return True

  for machine_backend in MachineBackend.get_all_instances():
    if not (
      (hasattr(machine_backend, "io") and _replace_io(machine_backend))
      or (hasattr(machine_backend, "interface") and _replace_io(machine_backend.interface))
    ):
      raise RuntimeError(f"Backend {machine_backend} not supported for validation")

  cr.start()


def end_validation():
  if cr is None:
    raise RuntimeError("Validation not started")
  cr.done()



"""Backend for BioTek Synergy HT plate reader.

The Synergy HT uses the same protocol as other BioTek plate readers,
with some commands disabled (e.g., slow mode).
"""

from typing import Optional

from pylabrobot.plate_reading.agilent.biotek_backend import BioTekPlateReaderBackend
from pylabrobot.resources import Plate


class SynergyHTBackend(BioTekPlateReaderBackend):
  """Backend for BioTek Synergy HT plate readers.

  The Synergy HT is compatible with the standard BioTek protocol,
  but does not support slow mode for open/close operations.
  """

  async def open(self, slow: bool = False) -> None:
    """Open the plate reader door / eject plate.

    Note: slow parameter is ignored on Synergy HT (not supported).
    """
    # Synergy HT doesn't support slow mode command (&), so skip it
    return await self.send_command("J")

  async def close(self, plate: Optional[Plate] = None, slow: bool = False) -> None:
    """Close the plate reader door / load plate.

    Note: slow parameter is ignored on Synergy HT (not supported).
    """
    # Synergy HT doesn't support slow mode command (&), so skip it
    self._plate = None
    if plate is not None:
      await self.set_plate(plate)
    return await self.send_command("A")



class Powder:
  """
  A class for powders.
  """

  def __init__(self, name: str) -> None:
    """
    Initializes a new instance of the Powder class.
    Args:
        name (str): The name of the powder.
    """
    self.name = name



from __future__ import annotations

import warnings
from abc import ABCMeta
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Sequence, Union, cast

from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.resources.tip import Tip, TipCreator
from pylabrobot.resources.tip_tracker import TipTracker, does_tip_tracking
from pylabrobot.serializer import deserialize

from .itemized_resource import ItemizedResource
from .resource import Resource


class TipSpot(Resource):
  """A tip spot, a location in a tip rack where there may or may not be a tip.

  Each tip produced by this spot is given a unique name based on the parent
  rack name, the spot name, and a per-spot counter (e.g. ``rack.A1#1``,
  ``rack.A1#2``), which is stored on the :class:`Tip` instance.
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    make_tip: TipCreator,
    size_z: float = 0,
    category: str = "tip_spot",
  ):
    """Initialize a tip spot.

    Args:
      name: the name of the tip spot.
      size_x: the size of the tip spot in the x direction.
      size_y: the size of the tip spot in the y direction.
      size_z: the size of the tip spot in the z direction.
      make_tip: a function that creates a tip for the tip spot.
      category: the category of the tip spot.
    """

    super().__init__(
      name,
      size_x=size_y,
      size_y=size_x,
      size_z=size_z,
      category=category,
    )
    self.tracker = TipTracker(thing="Tip spot")
    self.parent: Optional["TipRack"] = None

    self._tip_counter: int = 0

    self._make_tip_func = make_tip

    self.tracker.register_callback(self._state_updated)

  def _get_next_tip_name(self) -> str:
    """Generate a unique name for the next tip originating from this spot."""

    name = f"{self.name}#{self._tip_counter}"
    self._tip_counter += 1
    return name

  def make_tip(self) -> Tip:
    """Create a new tip instance for this spot and assign it a unique name."""

    # use introspection to see if _make_tip_func has a name parameter
    if "name" in self._make_tip_func.__code__.co_varnames:
      tip = self._make_tip_func(self._get_next_tip_name())
    else:
      warnings.warn(
        "The make_tip function should accept a 'name' parameter to assign unique names to tips.",
        DeprecationWarning,
      )
      tip = self._make_tip_func()  # type: ignore # ignore type check for deprecated behavior
      if getattr(tip, "name", None) is None:
        tip.name = self._get_next_tip_name()
        if hasattr(tip, "tracker"):
          tip.tracker.thing = tip.name

    return tip

  def get_tip(self) -> Tip:
    """Get a tip from the tip spot."""

    # Tracker will raise an error if there is no tip. We spawn a new tip if tip tracking is disabled
    tracks = does_tip_tracking() and not self.tracker.is_disabled
    if not self.tracker.has_tip and not tracks:
      self.tracker.add_tip(self.make_tip(), origin=self)

    return self.tracker.get_tip()

  def has_tip(self) -> bool:
    """Check if the tip spot has a tip."""
    return self.tracker.has_tip

  def empty(self) -> None:
    """Empty the tip spot."""
    self.tracker.remove_tip()

  def serialize(self) -> dict:
    """Serialize the tip spot."""
    return {
      **super().serialize(),
      "prototype_tip": self.make_tip().serialize(),
    }

  @classmethod
  def deserialize(cls, data: dict, allow_marshal: bool = False) -> TipSpot:
    """Deserialize a tip spot."""
    tip_data = data["prototype_tip"]

    def make_tip(name: str) -> Tip:
      tip_data_with_name = {**tip_data, "name": name}
      return cast(Tip, deserialize(tip_data_with_name, allow_marshal=allow_marshal))

    return cls(
      name=data["name"],
      size_x=data["size_x"],
      size_y=data["size_y"],
      size_z=data["size_z"],
      make_tip=make_tip,
      category=data.get("category", "tip_spot"),
    )

  def serialize_state(self) -> Dict[str, Any]:
    return self.tracker.serialize()

  def load_state(self, state: Dict[str, Any]):
    self.tracker.load_state(state)

  def get_identifier(self) -> str:
    """Get the (canonical) identifier, like `"A1"` of the tip spot in the parent tip rack. If the
    tip spot not in a tip rack, this will raise a ValueError."""

    if self.parent is None or not isinstance(self.parent, TipRack):
      raise ValueError("TipSpot must be in a tip rack to get its identifier.")

    return self.parent.get_child_identifier(self)


class TipRack(ItemizedResource[TipSpot], metaclass=ABCMeta):
  """Tip rack for disposable tips."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    ordered_items: Optional[Dict[str, TipSpot]] = None,
    ordering: Optional[OrderedDict[str, str]] = None,
    category: str = "tip_rack",
    model: Optional[str] = None,
    with_tips: bool = True,
  ):
    super().__init__(
      name,
      size_x,
      size_y,
      size_z,
      ordered_items=ordered_items,
      ordering=ordering,
      category=category,
      model=model,
    )

    if ordered_items is not None and len(ordered_items) > 0:
      if with_tips:
        self.fill()
      else:
        self.empty()

  def __repr__(self) -> str:
    return (
      f"{self.__class__.__name__}(name={self.name!r}, size_x={self._size_x}, "
      f"size_y={self._size_y}, size_z={self._size_z}, location={self.location})"
    )

  @staticmethod
  def _occupied_func(item: TipSpot):
    return "V" if item.has_tip() else "-"

  def get_tip(self, identifier: Union[str, int]) -> Tip:
    """Get the item with the given identifier.

    See :meth:`~.get_item` for more information.
    """

    return super().get_item(identifier).get_tip()

  def get_tips(self, identifier: Union[str, Sequence[int]]) -> List[Tip]:
    """Get the tips with the given identifier.

    See :meth:`~.get_items` for more information.
    """

    return [ts.get_tip() for ts in super().get_items(identifier)]

  def set_tip_state(self, tips: Union[List[bool], Dict[str, bool]]) -> None:
    """Set the initial tip tracking state of all tips in this tip rack.

    Args:
      tips: either a list of booleans (using integer indexing) or a dictionary of booleans (using
        string indexing) for whether each tip should be filled or empty.

    Examples:
      Filling the right half of a 96-well tip rack:

      >>> tip_rack.set_tip_state([[True] * 6 + [False] * 6] * 8)
    """

    should_have: Dict[Union[int, str], bool] = {}
    if isinstance(tips, list):
      for i, tip in enumerate(tips):
        should_have[i] = tip
    else:
      should_have = cast(Dict[Union[int, str], bool], tips)  # type?

    for identifier, should_have_tip in should_have.items():
      if should_have_tip and not self.get_item(identifier).has_tip():
        spot = self.get_item(identifier)
        spot.tracker.add_tip(spot.make_tip(), origin=spot, commit=True)
      elif not should_have_tip and self.get_item(identifier).has_tip():
        self.get_item(identifier).tracker.remove_tip(commit=True)

  def disable_tip_trackers(self) -> None:
    """Disable tip tracking for all tips in this tip rack."""
    for item in self.get_all_items():
      item.tracker.disable()

  def enable_tip_trackers(self) -> None:
    """Enable tip tracking for all tips in this tip rack."""
    for item in self.get_all_items():
      item.tracker.enable()

  def empty(self):
    """Empty the tip rack. This is useful when tip tracking is enabled and you are modifying
    the state manually (without the robot)."""
    self.set_tip_state([False] * self.num_items)

  def fill(self):
    """Fill the tip rack. This is useful when tip tracking is enabled and you are modifying
    the state manually (without the robot)."""
    self.set_tip_state([True] * self.num_items)

  def get_all_tips(self) -> List[Tip]:
    """Get all tips in the tip rack."""
    return [ts.get_tip() for ts in self.get_all_items()]


class NestedTipRack(TipRack):
  """A nested tip rack."""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    stacking_z_height: float,
    ordered_items: Optional[Dict[str, TipSpot]] = None,
    ordering: Optional[OrderedDict[str, str]] = None,
    category: str = "tip_rack",
    model: Optional[str] = None,
    with_tips: bool = True,
  ):
    # Call the superclass constructor
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      ordered_items=ordered_items,
      ordering=ordering,
      category=category,
      model=model,
      with_tips=with_tips,
    )

    self.stacking_z_height = stacking_z_height

  def __repr__(self) -> str:
    return (
      f"{self.__class__.__name__}(name={self.name!r}, size_x={self._size_x}, "
      f"size_y={self._size_y}, size_z={self._size_z}, "
      f"stacking_z_height={self.stacking_z_height}, location={self.location})"
    )

  def assign_child_resource(
    self,
    resource: Resource,
    location: Optional[Coordinate] = None,
    reassign: bool = True,
  ):
    if isinstance(resource, NestedTipRack):
      location = location or Coordinate(0, 0, self.stacking_z_height)
    else:
      assert location is not None, (
        "Location must be specified if " + "resource is not a NestedTipRack."
      )
    return super().assign_child_resource(resource, location=location, reassign=reassign)



import enum


class Liquid(enum.Enum):
  """A type of liquid (eg water, ethanol, etc.).

  Backends use this information to determine optimal parameters for aspirating and dispensing. In
  software like VENUS and EvoWare, this information is part of "Liquid Classes". In PyLabRobot,
  liquid classes are simply groups of parameters passed to lh.
  """

  ACETONITRIL80WATER20 = "Acetonitril/Water 80:20"  # TODO: need a better way to represent this.
  WATER = "Water"
  ETHANOL = "Ethanol 100%"
  GLYCERIN = "Glycerin"
  GLYCERIN80 = "Glycerin 80%"  # TODO: need a better way to represent this.
  DMSO = "DMSO"
  PLASMA = "Plasma 100%"
  SERUM = "Serum 100%"
  ACETONITRILE = "Acetonitrile 100%"
  DIMETHYLSULFOXID = "Dimethylsulfoxid 100%"
  BLOOD = "Blood"
  BRAINHOMOGENATE = "Brain-Homogenate"
  CHLOROFORM = "Chloroform 100%"
  METHANOL = "Methanol 100%"
  OCTANOL = "Octanol 100%"
  DNA_TRIS_EDTA = "100µg/ml DNA in Tris-EDTA Puffer"  # TODO: need a better way to represent this.
  PBS_BUFFER = "PBS Buffer"
  METHANOL70WATER030 = "Methanol/Water 70:30"  # TODO: need a better way to represent this.

  @staticmethod
  def from_str(s: str) -> "Liquid":
    """Some liquid classes have more than one name. This is a little Hamilton specific, will
    probably refactor in the future."""

    s = s.strip()

    if s.endswith(" for aliquot"):
      s = s[: -len(" for aliquot")]  # I don't think this is needed.

    if s in {"EtOH", "Ethanol", "EtOH 100%"}:
      s = "Ethanol 100%"
    elif s == "Serum":
      s = "Serum 100%"
    elif s in {"Acetonitril", "Acetonitril 100%", "Acetonitrile"}:
      s = "Acetonitrile 100%"
    elif s == "Blood (completely)":
      s = "Blood"
    elif s == "Glycerin80":
      s = "Glycerin 80%"
    elif s in {"Water", "SysFlWater"}:
      s = "Water"  # TODO: ?
    elif s == "Plasma":
      s = "Plasma 100%"
    elif s == "DMSO":
      s = "DMSO"

    return Liquid(s)



from .base_control import BaseControl
from .camera_control import CameraControl
from .config_control import ConfigControl
from .data_control import DataControl
from .injector_control import InjectorControl
from .measurement_control import MeasurementControl
from .movement_control import MovementControl
from .optics_control import OpticsControl
from .plate_transport_control import PlateControl
from .sensor_control import SensorControl
from .system_control import SystemControl



from typing import Optional

from pylabrobot.machines.machine import Machine
from pylabrobot.resources.coordinate import Coordinate
from pylabrobot.shaking import Shaker
from pylabrobot.temperature_controlling import TemperatureController

from .backend import HeaterShakerBackend


class HeaterShaker(TemperatureController, Shaker):
  """A heating and shaking machine"""

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    backend: HeaterShakerBackend,
    child_location: Coordinate,
    category: str = "heating_shaking",
    model: Optional[str] = None,
  ):
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      backend=backend,
      child_location=child_location,
      category=category,
      model=model,
    )
    self.backend: HeaterShakerBackend = backend  # fix type

  async def stop(self):
    await self.deactivate()
    await self.stop_shaking()
    await Machine.stop(self)



from pylabrobot.sealing.a4s_backend import A4SBackend
from pylabrobot.sealing.sealer import Sealer


def a4s(port: str) -> Sealer:
  # https://web.azenta.com/hubfs/azenta-files/resources/tech-drawings/TD-automated-roll-heat-sealer.pdf
  # 222 x 500 x 276 mm
  return Sealer(
    backend=A4SBackend(port=port),
  )



import itertools
import tempfile
import unittest
import unittest.mock
from typing import Any, List, Union, cast
from unittest.mock import PropertyMock

import pytest

from pylabrobot.liquid_handling.backends.backend import LiquidHandlerBackend
from pylabrobot.liquid_handling.backends.chatterbox import LiquidHandlerChatterboxBackend
from pylabrobot.liquid_handling.errors import ChannelizedError
from pylabrobot.liquid_handling.strictness import (
  Strictness,
  set_strictness,
)
from pylabrobot.liquid_handling.utils import get_tight_single_resource_liquid_op_offsets
from pylabrobot.resources import (
  PLT_CAR_L5AC_A00,
  TIP_CAR_480_A00,
  Container,
  Coordinate,
  Cor_96_wellplate_360ul_Fb,
  Deck,
  Lid,
  Plate,
  ResourceNotFoundError,
  ResourceStack,
  TipRack,
  nest_1_troughplate_195000uL_Vb,
  no_tip_tracking,
  set_tip_tracking,
)
from pylabrobot.resources.carrier import PlateHolder
from pylabrobot.resources.errors import (
  HasTipError,
  NoTipError,
)
from pylabrobot.resources.hamilton import (
  STARLetDeck,
  hamilton_96_tiprack_300uL_filter,
  hamilton_96_tiprack_1000uL_filter,
)
from pylabrobot.resources.revvity.plates import Revvity_384_wellplate_28ul_Ub
from pylabrobot.resources.utils import create_ordered_items_2d
from pylabrobot.resources.volume_tracker import (
  set_volume_tracking,
)
from pylabrobot.resources.well import Well
from pylabrobot.serializer import serialize

from .liquid_handler import LiquidHandler
from .standard import (
  Drop,
  DropTipRack,
  GripDirection,
  MultiHeadAspirationPlate,
  MultiHeadDispensePlate,
  Pickup,
  ResourcePickup,
  SingleChannelAspiration,
  SingleChannelDispense,
)


def _create_mock_backend(num_channels: int = 8):
  """Create a mock LiquidHandlerBackend with the specified number of channels."""
  mock = unittest.mock.create_autospec(LiquidHandlerBackend, instance=True)
  type(mock).num_channels = PropertyMock(return_value=num_channels)
  type(mock).num_arms = PropertyMock(return_value=1)
  type(mock).head96_installed = PropertyMock(return_value=True)
  mock.can_pick_up_tip.return_value = True
  return mock


def _make_asp(
  r: Container,
  vol: float,
  tip: Any,
  offset: Coordinate = Coordinate.zero(),
) -> SingleChannelAspiration:
  return SingleChannelAspiration(
    resource=r,
    volume=vol,
    tip=tip,
    offset=offset,
    flow_rate=None,
    liquid_height=None,
    blow_out_air_volume=None,
    mix=None,
  )


def _make_disp(
  r: Container,
  vol: float,
  tip: Any,
  offset: Coordinate = Coordinate.zero(),
) -> SingleChannelDispense:
  return SingleChannelDispense(
    resource=r,
    volume=vol,
    tip=tip,
    offset=offset,
    flow_rate=None,
    liquid_height=None,
    blow_out_air_volume=None,
    mix=None,
  )


class TestLiquidHandlerLayout(unittest.IsolatedAsyncioTestCase):
  def setUp(self):
    self.backend = _create_mock_backend(num_channels=8)
    self.deck = STARLetDeck()
    self.lh = LiquidHandler(self.backend, deck=self.deck)

  def test_resource_assignment(self):
    tip_car = TIP_CAR_480_A00(name="tip_carrier")
    tip_car[0] = hamilton_96_tiprack_300uL_filter(name="tip_rack_01")
    tip_car[1] = hamilton_96_tiprack_300uL_filter(name="tip_rack_02")
    tip_car[3] = hamilton_96_tiprack_1000uL_filter("tip_rack_04")

    plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    plt_car[0] = Cor_96_wellplate_360ul_Fb(name="aspiration plate")
    plt_car[2] = Cor_96_wellplate_360ul_Fb(name="dispense plate")

    self.deck.assign_child_resource(tip_car, rails=1)
    self.deck.assign_child_resource(plt_car, rails=21)

    # Test placing a carrier at a location where another carrier is located.
    with self.assertRaises(ValueError):
      dbl_plt_car_1 = PLT_CAR_L5AC_A00(name="double placed carrier 1")
      self.deck.assign_child_resource(dbl_plt_car_1, rails=1)

    with self.assertRaises(ValueError):
      dbl_plt_car_2 = PLT_CAR_L5AC_A00(name="double placed carrier 2")
      self.deck.assign_child_resource(dbl_plt_car_2, rails=2)

    with self.assertRaises(ValueError):
      dbl_plt_car_3 = PLT_CAR_L5AC_A00(name="double placed carrier 3")
      self.deck.assign_child_resource(dbl_plt_car_3, rails=20)

    # Test invalid rails.
    with self.assertRaises(ValueError):
      self.deck.assign_child_resource(plt_car, rails=-1)
    with self.assertRaises(ValueError):
      self.deck.assign_child_resource(plt_car, rails=42)
    with self.assertRaises(ValueError):
      self.deck.assign_child_resource(plt_car, rails=27)

  def test_get_resource(self):
    tip_car = TIP_CAR_480_A00(name="tip_carrier")
    tip_car[0] = hamilton_96_tiprack_300uL_filter(name="tip_rack_01")
    plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    plt_car[0] = Cor_96_wellplate_360ul_Fb(name="aspiration plate")
    self.deck.assign_child_resource(tip_car, rails=1)
    self.deck.assign_child_resource(plt_car, rails=10)

    # Get resource.
    self.assertEqual(self.lh.deck.get_resource("tip_carrier").name, "tip_carrier")
    self.assertEqual(self.lh.deck.get_resource("plate carrier").name, "plate carrier")

    # Get subresource.
    self.assertEqual(self.lh.deck.get_resource("tip_rack_01").name, "tip_rack_01")
    self.assertEqual(
      self.lh.deck.get_resource("aspiration plate").name,
      "aspiration plate",
    )

    # Get unknown resource.
    with self.assertRaises(ResourceNotFoundError):
      self.lh.deck.get_resource("unknown resource")

  def test_name_parameter(self):
    # Default name is derived from deck name
    deck = STARLetDeck()
    lh = LiquidHandler(_create_mock_backend(), deck=deck)
    self.assertEqual(lh.name, f"lh_{deck.name}")

    # Custom name
    deck2 = STARLetDeck()
    lh2 = LiquidHandler(_create_mock_backend(), deck=deck2, name="my_liquid_handler")
    self.assertEqual(lh2.name, "my_liquid_handler")

  def test_subcoordinates(self):
    tip_car = TIP_CAR_480_A00(name="tip_carrier")
    tip_car[0] = hamilton_96_tiprack_300uL_filter(name="tip_rack_01")
    tip_car[3] = hamilton_96_tiprack_1000uL_filter(name="tip_rack_04")
    plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    plt_car[0] = Cor_96_wellplate_360ul_Fb(name="aspiration plate")
    plt_car[2] = Cor_96_wellplate_360ul_Fb(name="dispense plate")
    self.deck.assign_child_resource(tip_car, rails=1)
    self.deck.assign_child_resource(plt_car, rails=10)

    # Rails 10 should be left of rails 1.
    self.assertGreater(
      self.lh.deck.get_resource("plate carrier").get_absolute_location().x,
      self.lh.deck.get_resource("tip_carrier").get_absolute_location().x,
    )

    # Verified with Hamilton Method Editor.
    # Carriers.
    self.assertEqual(
      self.lh.deck.get_resource("tip_carrier").get_absolute_location(),
      Coordinate(100.0, 63.0, 100.0),
    )
    self.assertEqual(
      self.lh.deck.get_resource("plate carrier").get_absolute_location(),
      Coordinate(302.5, 63.0, 100.0),
    )

    # Subresources.
    self.assertEqual(
      cast(TipRack, self.lh.deck.get_resource("tip_rack_01")).get_item("A1").get_absolute_location()
      + cast(TipRack, self.lh.deck.get_resource("tip_rack_01")).get_item("A1").center(),
      Coordinate(117.900, 145.800, 164.450),
    )
    self.assertEqual(
      cast(TipRack, self.lh.deck.get_resource("tip_rack_04")).get_item("A1").get_absolute_location()
      + cast(TipRack, self.lh.deck.get_resource("tip_rack_04")).get_item("A1").center(),
      Coordinate(117.900, 433.800, 131.450),
    )

    self.assertEqual(
      cast(Plate, self.lh.deck.get_resource("aspiration plate"))
      .get_item("A1")
      .get_absolute_location()
      + cast(Plate, self.lh.deck.get_resource("aspiration plate")).get_item("A1").center(),
      Coordinate(x=320.8, y=145.7, z=186.15),
    )

  def test_illegal_subresource_assignment_before(self):
    # Test assigning subresource with the same name as another resource in another carrier. This
    # should raise an ValueError when the carrier is assigned to the liquid handler.
    tip_car = TIP_CAR_480_A00(name="tip_carrier")
    tip_car[0] = hamilton_96_tiprack_300uL_filter(name="sub")
    plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    plt_car[0] = Cor_96_wellplate_360ul_Fb(name="sub")
    self.deck.assign_child_resource(tip_car, rails=1)
    with self.assertRaises(ValueError):
      self.deck.assign_child_resource(plt_car, rails=10)

  def test_illegal_subresource_assignment_after(self):
    # Test assigning subresource with the same name as another resource in another carrier, after
    # the carrier has been assigned. This should raise an error.
    tip_car = TIP_CAR_480_A00(name="tip_carrier")
    tip_car[0] = hamilton_96_tiprack_300uL_filter(name="sub")
    plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    plt_car[0] = Cor_96_wellplate_360ul_Fb(name="ok")
    self.deck.assign_child_resource(tip_car, rails=1)
    self.deck.assign_child_resource(plt_car, rails=10)
    with self.assertRaises(ValueError):
      plt_car[1] = Cor_96_wellplate_360ul_Fb(name="sub")

  async def test_move_plate_to_site(self):
    plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    plt_car[0] = plate = Cor_96_wellplate_360ul_Fb(name="plate")
    self.deck.assign_child_resource(plt_car, rails=21)

    await self.lh.move_plate(plate, plt_car[2])
    self.assertIsNotNone(plt_car[2].resource)
    self.assertIsNone(plt_car[0].resource)
    self.assertEqual(plt_car[2].resource, self.lh.deck.get_resource("plate"))
    self.assertEqual(
      plate.get_item("A1").get_absolute_location() + plate.get_item("A1").center(),
      Coordinate(x=568.3, y=337.7, z=186.15),
    )

  async def test_move_plate_free(self):
    plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    plt_car[0] = plate = Cor_96_wellplate_360ul_Fb(name="plate")
    self.deck.assign_child_resource(plt_car, rails=1)

    await self.lh.move_plate(plate, Coordinate(1000, 1000, 1000))
    self.assertIsNotNone(self.lh.deck.get_resource("plate"))
    self.assertIsNone(plt_car[0].resource)
    self.assertEqual(plate.get_absolute_location(), Coordinate(1000, 1000, 1000))

  async def test_move_lid(self):
    plate = Plate("plate", size_x=100, size_y=100, size_z=15, ordered_items={})
    self.deck.assign_child_resource(plate, location=Coordinate(0, 0, 100))
    lid_height = 10
    lid = Lid(
      name="lid",
      size_x=plate.get_absolute_size_x(),
      size_y=plate.get_absolute_size_y(),
      size_z=lid_height,
      nesting_z_height=lid_height,
    )
    self.deck.assign_child_resource(lid, location=Coordinate(100, 100, 200))

    assert plate.get_absolute_location().x != lid.get_absolute_location().x
    assert plate.get_absolute_location().y != lid.get_absolute_location().y
    assert (
      plate.get_absolute_location().z + plate.get_absolute_size_z() - lid_height
      != lid.get_absolute_location().z
    )

    await self.lh.move_lid(lid, plate)

    assert plate.get_absolute_location().x == lid.get_absolute_location().x
    assert plate.get_absolute_location().y == lid.get_absolute_location().y
    assert (
      plate.get_absolute_location().z + plate.get_absolute_size_z() - lid_height
      == lid.get_absolute_location().z
    )

  async def test_move_plate_onto_resource_stack_with_lid(self):
    plate = Plate("plate", size_x=100, size_y=100, size_z=15, ordered_items={})
    lid = Lid(
      name="lid",
      size_x=plate.get_absolute_size_x(),
      size_y=plate.get_absolute_size_y(),
      size_z=10,
      nesting_z_height=4,
    )

    stack = ResourceStack("stack", direction="z")
    self.deck.assign_child_resource(stack, location=Coordinate(100, 100, 0))

    await self.lh.move_plate(plate, stack)
    await self.lh.move_lid(lid, plate)

    assert plate.location is not None
    self.assertEqual(plate.location.z, 0)
    assert lid.location is not None
    self.assertEqual(lid.location.z, 11)
    self.assertEqual(plate.lid, lid)
    self.assertEqual(stack.get_absolute_size_z(), 21)

  async def test_move_plate_onto_resource_stack_with_plate(self):
    plate1 = Plate("plate1", size_x=100, size_y=100, size_z=15, ordered_items={})
    plate2 = Plate("plate2", size_x=100, size_y=100, size_z=15, ordered_items={})

    stack = ResourceStack("stack", direction="z")

    self.deck.assign_child_resource(stack, location=Coordinate(100, 100, 0))
    await self.lh.move_plate(plate1, stack)
    await self.lh.move_plate(plate2, stack)

    assert plate1.location is not None and plate2.location is not None
    self.assertEqual(plate1.location.z, 0)
    self.assertEqual(plate2.location.z, 15)
    self.assertEqual(stack.get_absolute_size_z(), 30)

  async def test_move_plate_rotation(self):
    rotations = [0, 180, 360]  # rotation wrt site before AND after move
    grip_directions = [
      (GripDirection.LEFT, GripDirection.RIGHT),
      (GripDirection.FRONT, GripDirection.BACK),
    ]
    sites: List[Union[ResourceStack, PlateHolder]] = [
      ResourceStack(name="stack", direction="z"),
      PlateHolder(
        name="site",
        size_x=100,
        size_y=100,
        size_z=15,
        pedestal_size_z=1,
      ),
    ]

    test_cases = itertools.product(sites, rotations, grip_directions)

    for site, rotation, (pickup_direction, drop_direction) in test_cases:
      with self.subTest(
        stack_type=site.__class__.__name__,
        rotation=rotation,
        pickup_direction=pickup_direction,
        drop_direction=drop_direction,
      ):
        self.deck.assign_child_resource(site, location=Coordinate(100, 100, 0))

        plate = Plate(
          "plate",
          size_x=200,
          size_y=100,
          size_z=15,
          ordered_items=create_ordered_items_2d(
            Well,
            num_items_x=1,
            num_items_y=1,
            dx=0,
            dy=0,
            dz=0,
            item_dx=10,
            item_dy=10,
            size_x=10,
            size_y=10,
            size_z=10,
          ),
        )
        plate.rotate(z=rotation)
        site.assign_child_resource(plate)
        original_center = plate.get_absolute_location(x="c", y="c", z="c")
        await self.lh.move_plate(
          plate,
          site,
          pickup_direction=pickup_direction,
          drop_direction=drop_direction,
        )
        new_center = plate.get_absolute_location(x="c", y="c", z="c")
        assert plate.rotation.z == (rotation + 180) % 360

        self.assertEqual(
          new_center,
          original_center,
          f"Center mismatch for {site.__class__.__name__}, rotation {rotation}, "
          f"pickup_direction {pickup_direction}, "
          f"drop_direction {drop_direction}",
        )
        plate.unassign()
        self.deck.unassign_child_resource(site)

  async def test_move_lid_rotation(self):
    rotations = [0, 90, 270, 360]
    grip_directions = [
      (GripDirection.LEFT, GripDirection.RIGHT),
      (GripDirection.FRONT, GripDirection.BACK),
    ]

    test_cases = itertools.product(rotations, grip_directions)

    plate = Plate("plate", size_x=200, size_y=100, size_z=15, ordered_items={})
    lid = Lid(
      name="lid",
      size_x=plate.get_absolute_size_x(),
      size_y=plate.get_absolute_size_y(),
      size_z=10,
      nesting_z_height=4,
    )
    self.deck.assign_child_resource(plate, location=Coordinate(100, 100, 0))
    for rot, (pickup_direction, drop_direction) in test_cases:
      with self.subTest(
        rotation=rot,
        pickup_direction=pickup_direction,
        drop_direction=drop_direction,
      ):
        plate.rotate(z=rot)
        plate.assign_child_resource(lid)
        original_center = lid.get_absolute_location(x="c", y="c", z="c")
        await self.lh.move_lid(
          lid,
          plate,
          pickup_direction=pickup_direction,
          drop_direction=drop_direction,
        )
        new_center = lid.get_absolute_location(x="c", y="c", z="c")
        self.assertEqual(
          new_center,
          original_center,
          f"Center mismatch for rotation {rot}, pickup_direction {pickup_direction}, "
          f"drop_direction {drop_direction}",
        )
        lid.unassign()
        # reset rotations
        plate.rotation.z = 0
        lid.rotation.z = 0

  def test_serialize(self):
    # Use a real backend for serialization test since mocks can't be deserialized
    backend = LiquidHandlerChatterboxBackend(num_channels=8)
    lh = LiquidHandler(backend, deck=STARLetDeck())

    serialized = lh.serialize()
    deserialized = LiquidHandler.deserialize(serialized)

    self.assertEqual(deserialized.deck, lh.deck)
    self.assertEqual(
      deserialized.backend.__class__.__name__,
      lh.backend.__class__.__name__,
    )


class TestLiquidHandlerCommands(unittest.IsolatedAsyncioTestCase):
  async def asyncSetUp(self):
    self.maxDiff = None

    self.backend = _create_mock_backend(num_channels=8)
    self.deck = STARLetDeck()
    self.lh = LiquidHandler(backend=self.backend, deck=self.deck)

    self.tip_rack = hamilton_96_tiprack_300uL_filter(name="tip_rack")
    self.plate = Cor_96_wellplate_360ul_Fb(name="plate")
    self.deck.assign_child_resource(self.tip_rack, location=Coordinate(0, 0, 0))
    self.deck.assign_child_resource(self.plate, location=Coordinate(100, 100, 0))
    await self.lh.setup()

  async def test_offsets_tips(self):
    tip_spot = self.tip_rack.get_item("A1")
    tip = tip_spot.get_tip()
    await self.lh.pick_up_tips([tip_spot], offsets=[Coordinate(x=1, y=1, z=1)])
    await self.lh.drop_tips([tip_spot], offsets=[Coordinate(x=1, y=1, z=1)])

    self.backend.pick_up_tips.assert_called_once_with(
      use_channels=[0],
      ops=[Pickup(tip_spot, tip=tip, offset=Coordinate(x=1, y=1, z=1))],
    )
    self.backend.drop_tips.assert_called_once_with(
      use_channels=[0],
      ops=[Drop(tip_spot, tip=tip, offset=Coordinate(x=1, y=1, z=1))],
    )

  async def test_default_offset_head96(self):
    self.lh.default_offset_head96 = Coordinate(1, 2, 3)

    await self.lh.pick_up_tips96(self.tip_rack)
    self.backend.pick_up_tips96.assert_called_once()
    call_kwargs = self.backend.pick_up_tips96.call_args.kwargs
    self.assertEqual(call_kwargs["pickup"].offset, Coordinate(1, 2, 3))
    self.backend.pick_up_tips96.reset_mock()

    # aspirate with extra offset; effective offset should be default + provided
    await self.lh.aspirate96(self.plate, volume=10, offset=Coordinate(1, 0, 0))
    self.backend.aspirate96.assert_called_once()
    call_kwargs = self.backend.aspirate96.call_args.kwargs
    self.assertEqual(call_kwargs["aspiration"].offset, Coordinate(2, 2, 3))
    self.backend.aspirate96.reset_mock()

    # dispense without providing offset uses default
    await self.lh.dispense96(self.plate, volume=10)
    self.backend.dispense96.assert_called_once()
    call_kwargs = self.backend.dispense96.call_args.kwargs
    self.assertEqual(call_kwargs["dispense"].offset, Coordinate(1, 2, 3))
    self.backend.dispense96.reset_mock()

    await self.lh.drop_tips96(self.tip_rack, offset=Coordinate(0, 1, 0))
    self.backend.drop_tips96.assert_called_once()
    call_kwargs = self.backend.drop_tips96.call_args.kwargs
    self.assertEqual(call_kwargs["drop"].offset, Coordinate(1, 3, 3))

  async def test_default_offset_head96_initializer(self):
    backend = _create_mock_backend(num_channels=8)
    deck = STARLetDeck()
    lh = LiquidHandler(
      backend=backend,
      deck=deck,
      default_offset_head96=Coordinate(1, 2, 3),
    )
    self.assertEqual(lh.default_offset_head96, Coordinate(1, 2, 3))

  async def test_default_offset_head96_serialization(self):
    # Use a real backend for serialization test since mocks can't be deserialized
    backend = LiquidHandlerChatterboxBackend(num_channels=8)
    lh = LiquidHandler(backend=backend, deck=STARLetDeck())
    lh.default_offset_head96 = Coordinate(1, 2, 3)
    data = lh.serialize()
    self.assertEqual(data["default_offset_head96"], serialize(Coordinate(1, 2, 3)))
    new_lh = LiquidHandler.deserialize(data)
    self.assertEqual(new_lh.default_offset_head96, Coordinate(1, 2, 3))

  async def test_with_use_channels(self):
    tip_spot = self.tip_rack.get_item("A1")
    tip = tip_spot.get_tip()
    with self.lh.use_channels([2]):
      await self.lh.pick_up_tips([tip_spot])
      await self.lh.drop_tips([tip_spot])

    self.backend.pick_up_tips.assert_called_once_with(
      use_channels=[2],
      ops=[Pickup(tip_spot, tip=tip, offset=Coordinate.zero())],
    )
    self.backend.drop_tips.assert_called_once_with(
      use_channels=[2],
      ops=[Drop(tip_spot, tip=tip, offset=Coordinate.zero())],
    )

  async def test_offsets_asp_disp(self):
    well = self.plate.get_item("A1")
    self.plate.get_item("A1").tracker.set_volume(10)
    t = self.tip_rack.get_item("A1").get_tip()
    self.lh.update_head_state({0: t})
    await self.lh.aspirate([well], vols=[10], offsets=[Coordinate(x=1, y=1, z=1)])
    await self.lh.dispense([well], vols=[10], offsets=[Coordinate(x=1, y=1, z=1)])

    self.backend.aspirate.assert_called_once_with(
      use_channels=[0],
      ops=[_make_asp(well, vol=10, offset=Coordinate(x=1, y=1, z=1), tip=t)],
    )
    self.backend.dispense.assert_called_once_with(
      use_channels=[0],
      ops=[_make_disp(well, vol=10, offset=Coordinate(x=1, y=1, z=1), tip=t)],
    )

  async def test_return_tips(self):
    tip_spot = self.tip_rack.get_item("A1")
    tip = tip_spot.get_tip()
    await self.lh.pick_up_tips([tip_spot])
    await self.lh.return_tips()

    self.backend.drop_tips.assert_called_once_with(
      use_channels=[0],
      ops=[Drop(tip_spot, tip=tip, offset=Coordinate.zero())],
    )

    with self.assertRaises(RuntimeError):
      await self.lh.return_tips()

  async def test_return_tips96(self):
    for i in range(96):
      assert not self.lh.head96[i].has_tip, f"Channel head {i} is not empty."
    await self.lh.pick_up_tips96(self.tip_rack)
    for i in range(96):
      assert self.lh.head96[i].has_tip, f"Channel head {i} is empty."
    await self.lh.return_tips96()
    for i in range(96):
      assert not self.lh.head96[i].has_tip, f"Channel head {i} is not empty."

    self.backend.drop_tips96.assert_called_once_with(
      drop=DropTipRack(resource=self.tip_rack, offset=Coordinate.zero())
    )

    with self.assertRaises(RuntimeError):
      await self.lh.return_tips()

  async def test_aspirate_dispense96(self):
    await self.lh.pick_up_tips96(self.tip_rack)
    await self.lh.aspirate96(self.plate, volume=10)
    await self.lh.dispense96(self.plate, 10)
    self.backend.dispense96.assert_called_with(
      dispense=MultiHeadDispensePlate(
        wells=self.plate.get_all_items(),
        offset=Coordinate.zero(),
        tips=[self.lh.head96[i].get_tip() for i in range(96)],
        volume=10,
        flow_rate=None,
        liquid_height=None,
        blow_out_air_volume=None,
        mix=None,
      )
    )

  async def test_dispense96_with_quadrant_well_list(self):
    plate_384 = Revvity_384_wellplate_28ul_Ub(name="plate_384")
    self.deck.assign_child_resource(plate_384, location=Coordinate(400, 100, 0))
    quadrant_wells = plate_384.get_quadrant("tl")

    await self.lh.pick_up_tips96(self.tip_rack)
    await self.lh.aspirate96(self.plate, volume=10)

    self.lh.backend.dispense96 = unittest.mock.AsyncMock()  # type: ignore
    await self.lh.dispense96(quadrant_wells, 10)
    self.lh.backend.dispense96.assert_called_with(  # type: ignore
      dispense=MultiHeadDispensePlate(
        wells=quadrant_wells,
        offset=Coordinate.zero(),
        tips=[self.lh.head96[i].get_tip() for i in range(96)],
        volume=10,
        flow_rate=None,
        liquid_height=None,
        blow_out_air_volume=None,
        mix=None,
      )
    )

  async def test_dispense96_well_list_mixed_parents(self):
    plate2 = Cor_96_wellplate_360ul_Fb(name="plate2")
    self.deck.assign_child_resource(plate2, location=Coordinate(400, 100, 0))
    mixed = self.plate.get_all_items()[:48] + plate2.get_all_items()[:48]
    await self.lh.pick_up_tips96(self.tip_rack)
    with self.assertRaises(ValueError):
      await self.lh.dispense96(mixed, 10)

  async def test_transfer(self):
    t = self.tip_rack.get_item("A1").get_tip()
    self.lh.update_head_state({0: t})

    # Simple transfer
    self.plate.get_item("A1").tracker.set_volume(10)
    await self.lh.transfer(self.plate.get_well("A1"), self.plate["A2"], source_vol=10)

    self.backend.aspirate.assert_called_once_with(
      use_channels=[0],
      ops=[_make_asp(self.plate.get_item("A1"), vol=10.0, tip=t)],
    )
    self.backend.dispense.assert_called_once_with(
      use_channels=[0],
      ops=[_make_disp(self.plate.get_item("A2"), vol=10.0, tip=t)],
    )
    self.backend.aspirate.reset_mock()
    self.backend.dispense.reset_mock()

    # Transfer to multiple wells
    self.plate.get_item("A1").tracker.set_volume(80)
    await self.lh.transfer(self.plate.get_well("A1"), self.plate["A1:H1"], source_vol=80)
    self.backend.aspirate.assert_called_once_with(
      use_channels=[0],
      ops=[_make_asp(self.plate.get_item("A1"), vol=80.0, tip=t)],
    )

    dispense_calls = self.backend.dispense.call_args_list
    expected_dispenses = [
      unittest.mock.call(use_channels=[0], ops=[_make_disp(well, vol=10.0, tip=t)])
      for well in self.plate["A1:H1"]
    ]
    self.assertEqual(dispense_calls, expected_dispenses)
    self.backend.aspirate.reset_mock()
    self.backend.dispense.reset_mock()

    # Transfer with ratios
    self.plate.get_item("A1").tracker.set_volume(60)
    await self.lh.transfer(
      self.plate.get_well("A1"),
      self.plate["B1:C1"],
      source_vol=60,
      ratios=[2, 1],
    )
    self.backend.aspirate.assert_called_once_with(
      use_channels=[0],
      ops=[_make_asp(self.plate.get_item("A1"), vol=60.0, tip=t)],
    )
    dispense_calls = self.backend.dispense.call_args_list
    expected_dispenses = [
      unittest.mock.call(use_channels=[0], ops=[_make_disp(well, vol=vol, tip=t)])
      for well, vol in zip(self.plate["B1:C1"], [40, 20])
    ]
    self.assertEqual(dispense_calls, expected_dispenses)
    self.backend.aspirate.reset_mock()
    self.backend.dispense.reset_mock()

    # Transfer with target_vols
    vols: List[float] = [3, 1, 4, 1, 5, 9, 6, 2]
    self.plate.get_item("A1").tracker.set_volume(sum(vols))
    await self.lh.transfer(self.plate.get_well("A1"), self.plate["A1:H1"], target_vols=vols)
    self.backend.aspirate.assert_called_once_with(
      use_channels=[0],
      ops=[_make_asp(self.plate.get_well("A1"), vol=sum(vols), tip=t)],
    )
    dispense_calls = self.backend.dispense.call_args_list
    expected_dispenses = [
      unittest.mock.call(use_channels=[0], ops=[_make_disp(well, vol=vol, tip=t)])
      for well, vol in zip(self.plate["A1:H1"], vols)
    ]
    self.assertEqual(dispense_calls, expected_dispenses)
    self.backend.aspirate.reset_mock()
    self.backend.dispense.reset_mock()

    # target_vols and source_vol specified
    with self.assertRaises(TypeError):
      await self.lh.transfer(
        self.plate.get_well("A1"),
        self.plate["A1:H1"],
        source_vol=100,
        target_vols=vols,
      )

    # target_vols and ratios specified
    with self.assertRaises(TypeError):
      await self.lh.transfer(
        self.plate.get_well("A1"),
        self.plate["A1:H1"],
        ratios=[1] * 8,
        target_vols=vols,
      )

  async def test_stamp(self):
    # Simple transfer
    await self.lh.pick_up_tips96(self.tip_rack)  # pick up tips first.
    await self.lh.stamp(self.plate, self.plate, volume=10)
    ts = self.tip_rack.get_all_tips()

    self.backend.aspirate96.assert_called_once_with(
      aspiration=MultiHeadAspirationPlate(
        wells=self.plate.get_all_items(),
        volume=10.0,
        tips=ts,
        offset=Coordinate.zero(),
        flow_rate=None,
        liquid_height=None,
        blow_out_air_volume=None,
        mix=None,
      )
    )
    self.backend.dispense96.assert_called_once_with(
      dispense=MultiHeadDispensePlate(
        wells=self.plate.get_all_items(),
        volume=10.0,
        tips=ts,
        offset=Coordinate.zero(),
        flow_rate=None,
        liquid_height=None,
        blow_out_air_volume=None,
        mix=None,
      )
    )

  async def test_tip_tracking_double_pickup(self):
    set_tip_tracking(enabled=True)
    await self.lh.pick_up_tips(self.tip_rack["A1"])
    with self.assertRaises(HasTipError):
      await self.lh.pick_up_tips(self.tip_rack["A2"])
    await self.lh.drop_tips(self.tip_rack["A1"])
    # pick_up_tips should work even after causing a HasTipError
    await self.lh.pick_up_tips(self.tip_rack["A2"])
    await self.lh.drop_tips(self.tip_rack["A2"])
    set_tip_tracking(enabled=False)

    self.lh.clear_head_state()
    with no_tip_tracking():
      await self.lh.pick_up_tips(self.tip_rack["A2"])

  async def test_tip_tracking_empty_drop(self):
    with self.assertRaises(NoTipError):
      await self.lh.drop_tips(self.tip_rack["A1"])

    await self.lh.pick_up_tips(self.tip_rack["A2"])
    set_tip_tracking(enabled=True)
    with self.assertRaises(HasTipError):
      await self.lh.drop_tips(self.tip_rack["A3"])
    set_tip_tracking(enabled=False)

  async def test_tip_tracking_empty_pickup(self):
    self.tip_rack.get_item("A1").empty()

    set_tip_tracking(enabled=True)
    with self.assertRaises(NoTipError):
      await self.lh.pick_up_tips(self.tip_rack["A1"])
    set_tip_tracking(enabled=False)

  async def test_get_mounted_tips(self):
    self.assertEqual(self.lh.get_mounted_tips(), [None] * 8)
    await self.lh.pick_up_tips(self.tip_rack["A1", "B1", "C1"])
    mounted = self.lh.get_mounted_tips()
    self.assertIsNotNone(self.tip_rack.get_item("A1").get_tip())
    self.assertIsNotNone(self.tip_rack.get_item("B1").get_tip())
    self.assertIsNotNone(self.tip_rack.get_item("C1").get_tip())
    self.assertIsNone(mounted[3])
    self.assertIsNone(mounted[4])
    self.assertIsNone(mounted[5])
    self.assertIsNone(mounted[6])
    self.assertIsNone(mounted[7])

  async def test_tip_tracking_full_spot(self):
    await self.lh.pick_up_tips(self.tip_rack["A1"])
    with self.assertRaises(HasTipError):
      set_tip_tracking(enabled=True)
      await self.lh.drop_tips(self.tip_rack["A2"])
      set_tip_tracking(enabled=False)

  async def test_tip_tracking_double_pickup_single_command(self):
    set_tip_tracking(enabled=True)
    with self.assertRaises(NoTipError):
      await self.lh.pick_up_tips(self.tip_rack["A1", "A1"])
    set_tip_tracking(enabled=False)

  async def test_discard_tips(self):
    tips = self.tip_rack.get_tips("A1:D1")
    await self.lh.pick_up_tips(self.tip_rack["A1", "B1", "C1", "D1"], use_channels=[0, 1, 3, 4])
    await self.lh.discard_tips()
    trash = self.deck.get_trash_area()
    offsets = get_tight_single_resource_liquid_op_offsets(trash, num_channels=4)

    # drop_tips is called twice: once for pick_up, so check the second call
    drop_tips_calls = self.backend.drop_tips.call_args_list
    self.assertEqual(len(drop_tips_calls), 1)
    self.assertEqual(
      drop_tips_calls[0],
      unittest.mock.call(
        use_channels=[0, 1, 3, 4],
        ops=[
          Drop(
            self.deck.get_trash_area(),
            tip=tips[3],
            offset=offsets[0],
          ),
          Drop(
            self.deck.get_trash_area(),
            tip=tips[2],
            offset=offsets[1],
          ),
          Drop(
            self.deck.get_trash_area(),
            tip=tips[1],
            offset=offsets[2],
          ),
          Drop(
            self.deck.get_trash_area(),
            tip=tips[0],
            offset=offsets[3],
          ),
        ],
      ),
    )

    # test tip tracking
    with self.assertRaises(RuntimeError):
      await self.lh.discard_tips()

  async def test_aspirate_with_lid(self):
    lid = Lid(
      "lid",
      size_x=self.plate.get_size_x(),
      size_y=self.plate.get_size_y(),
      size_z=10,
      nesting_z_height=self.plate.get_size_z(),
    )
    self.plate.assign_child_resource(lid)
    well = self.plate.get_item("A1")
    t = self.tip_rack.get_item("A1").get_tip()
    self.lh.update_head_state({0: t})
    with self.assertRaises(ValueError):
      await self.lh.aspirate([well], vols=[10])

  @pytest.mark.filterwarnings("ignore:Extra arguments to backend")
  async def test_strictness(self):
    # Create a mock backend with a custom pick_up_tips that checks arguments
    async def custom_pick_up_tips(ops, use_channels, non_default, default=True):
      assert non_default == default

    self.backend = _create_mock_backend(num_channels=16)
    self.backend.pick_up_tips = custom_pick_up_tips
    self.lh = LiquidHandler(self.backend, deck=self.deck)
    await self.lh.setup()

    with no_tip_tracking():
      set_strictness(Strictness.IGNORE)
      await self.lh.pick_up_tips(self.tip_rack["A1"], non_default=True)
      await self.lh.pick_up_tips(
        self.tip_rack["A1"],
        use_channels=[1],
        non_default=True,
        does_not_exist=True,
      )
      with self.assertRaises(TypeError):  # missing non_default
        await self.lh.pick_up_tips(self.tip_rack["A1"], use_channels=[2])

      set_strictness(Strictness.WARN)
      await self.lh.pick_up_tips(self.tip_rack["A1"], non_default=True, use_channels=[3])
      with self.assertWarns(UserWarning):  # extra kwargs should warn
        await self.lh.pick_up_tips(
          self.tip_rack["A1"],
          use_channels=[4],
          non_default=True,
          does_not_exist=True,
        )
      self.lh.clear_head_state()
      # We override default to False, so this should raise an assertion error. To test whether
      # overriding default to True works.
      with self.assertRaises(AssertionError):
        await self.lh.pick_up_tips(
          self.tip_rack["A1"],
          use_channels=[4],
          non_default=True,
          does_not_exist=True,
          default=False,
        )
      with self.assertRaises(TypeError):  # missing non_default
        await self.lh.pick_up_tips(self.tip_rack["A1"], use_channels=[5])

      set_strictness(Strictness.STRICT)
      self.lh.clear_head_state()
      await self.lh.pick_up_tips(self.tip_rack["A1"], non_default=True, use_channels=[6])
      with self.assertRaises(TypeError):  # cannot have extra kwargs
        await self.lh.pick_up_tips(
          self.tip_rack["A1"],
          use_channels=[7],
          non_default=True,
          does_not_exist=True,
        )
      with self.assertRaises(TypeError):  # missing non_default
        await self.lh.pick_up_tips(self.tip_rack["A1"], use_channels=[8])

      set_strictness(Strictness.WARN)

  async def test_save_state(self):
    set_volume_tracking(enabled=True)

    # set and save the state
    self.plate.get_item("A2").tracker.set_volume(10)
    state_filename = tempfile.mktemp()
    self.lh.deck.save_state_to_file(fn=state_filename)

    # save the deck
    deck_filename = tempfile.mktemp()
    self.lh.deck.save(fn=deck_filename)

    # create a new liquid handler, load the state and the deck
    backend2 = _create_mock_backend(num_channels=8)
    lh2 = LiquidHandler(backend2, deck=STARLetDeck())
    lh2.deck = Deck.load_from_json_file(json_file=deck_filename)
    lh2.deck.load_state_from_file(fn=state_filename)

    # assert that the state is the same
    well_a1 = lh2.deck.get_resource("plate").get_item("A1")  # type: ignore
    self.assertEqual(well_a1.tracker.volume, 0)
    well_a2 = lh2.deck.get_resource("plate").get_item("A2")  # type: ignore
    self.assertEqual(well_a2.tracker.volume, 10)

    set_volume_tracking(enabled=False)

  async def test_pick_up_tips96_incomplete_rack(self):
    set_tip_tracking(enabled=True)

    # Test that picking up tips from an incomplete rack works
    self.tip_rack.fill()
    self.tip_rack.get_item("A1").tracker.remove_tip()

    await self.lh.pick_up_tips96(self.tip_rack)

    # Check that the tips were picked up correctly
    self.assertFalse(self.lh.head96[0].has_tip)
    for i in range(1, 96):
      self.assertTrue(self.lh.head96[i].has_tip)

    set_tip_tracking(enabled=False)


class TestLiquidHandlerVolumeTracking(unittest.IsolatedAsyncioTestCase):
  async def asyncSetUp(self):
    self.backend = _create_mock_backend(num_channels=8)
    self.deck = STARLetDeck()
    self.lh = LiquidHandler(backend=self.backend, deck=self.deck)
    self.tip_rack = hamilton_96_tiprack_300uL_filter(name="tip_rack")
    self.deck.assign_child_resource(self.tip_rack, location=Coordinate(0, 0, 0))
    self.plate = Cor_96_wellplate_360ul_Fb(name="plate")
    self.deck.assign_child_resource(self.plate, location=Coordinate(100, 100, 0))
    self.single_well_plate = nest_1_troughplate_195000uL_Vb(name="single_well_plate")
    self.deck.assign_child_resource(self.single_well_plate, location=Coordinate(300, 100, 0))
    await self.lh.setup()
    set_volume_tracking(enabled=True)

  async def asyncTearDown(self):
    set_volume_tracking(enabled=False)

  async def test_dispense_with_volume_tracking(self):
    well = self.plate.get_item("A1")
    await self.lh.pick_up_tips(self.tip_rack["A1"])
    well.tracker.set_volume(10)
    await self.lh.aspirate([well], vols=[10])
    await self.lh.dispense([well], vols=[10])
    self.assertEqual(well.tracker.volume, 10)

  async def test_mix_volume_tracking(self):
    for i in range(8):
      self.plate.get_item(i).set_volume(55)

    await self.lh.pick_up_tips(self.tip_rack[0:8])
    for _ in range(10):
      await self.lh.aspirate(self.plate[0:8], vols=[45] * 8)
      await self.lh.dispense(self.plate[0:8], vols=[45] * 8)

  async def test_channel_1_liquid_tracking(self):
    self.plate.get_item("A1").tracker.set_volume(10)
    with self.lh.use_channels([1]):
      await self.lh.pick_up_tips(self.tip_rack["A1"])
      await self.lh.aspirate([self.plate.get_item("A1")], vols=[10])
      await self.lh.dispense([self.plate.get_item("A2")], vols=[10])

  async def test_dispense_fails(self):
    well = self.plate.get_item("A1")
    await self.lh.pick_up_tips(self.tip_rack["A1"])

    async def error_func(*args, **kwargs):
      raise ChannelizedError(errors={0: Exception("This is an error")})

    self.backend.dispense = error_func  # type: ignore
    well.tracker.set_volume(200)

    await self.lh.aspirate([well], vols=[200])
    assert self.lh.head[0].get_tip().tracker.get_used_volume() == 200
    with self.assertRaises(ChannelizedError):
      await self.lh.dispense([well], vols=[60])
    # test volume doesn't change on failed dispense
    assert self.lh.head[0].get_tip().tracker.get_used_volume() == 200

  async def test_96_head_volume_tracking_multi_container(self):
    for item in self.plate.get_all_items():
      item.tracker.set_volume(10)
    await self.lh.pick_up_tips96(self.tip_rack)
    await self.lh.aspirate96(self.plate, volume=10)
    for i in range(96):
      self.assertEqual(self.lh.head96[i].get_tip().tracker.get_used_volume(), 10)
      self.plate.get_item(i).tracker.get_used_volume() == 0
    await self.lh.dispense96(self.plate, volume=10)
    for i in range(96):
      self.assertEqual(self.lh.head96[i].get_tip().tracker.get_used_volume(), 0)
      self.plate.get_item(i).tracker.get_used_volume() == 10
    await self.lh.return_tips96()

  async def test_96_head_volume_tracking_single_container(self):
    well = self.single_well_plate.get_item(0)
    well.tracker.set_volume(10 * 96)
    await self.lh.pick_up_tips96(self.tip_rack)

    await self.lh.aspirate96(self.single_well_plate, volume=10)
    assert all(self.lh.head96[i].get_tip().tracker.get_used_volume() == 10 for i in range(96))
    assert all(self.lh.head96[i].get_tip().tracker.volume == 10 for i in range(96))
    assert well.tracker.get_used_volume() == 0

    await self.lh.dispense96(self.single_well_plate, volume=10)
    assert all(self.lh.head96[i].get_tip().tracker.get_used_volume() == 0 for i in range(96))
    assert all(self.lh.head96[i].get_tip().tracker.volume == 0 for i in range(96))
    assert well.tracker.get_used_volume() == 10 * 96

    await self.lh.return_tips96()

  async def test_96_head_volume_tracking_well_list(self):
    plate_384 = Revvity_384_wellplate_28ul_Ub(name="plate_384")
    self.deck.assign_child_resource(plate_384, location=Coordinate(600, 100, 0))
    quadrant_wells = plate_384.get_quadrant("tl")
    for well in quadrant_wells:
      well.tracker.set_volume(10)

    await self.lh.pick_up_tips96(self.tip_rack)
    await self.lh.aspirate96(quadrant_wells, volume=10)
    assert all(self.lh.head96[i].get_tip().tracker.get_used_volume() == 10 for i in range(96))
    assert all(w.tracker.get_used_volume() == 0 for w in quadrant_wells)

    await self.lh.dispense96(quadrant_wells, volume=10)
    assert all(self.lh.head96[i].get_tip().tracker.get_used_volume() == 0 for i in range(96))
    assert all(w.tracker.get_used_volume() == 10 for w in quadrant_wells)
    await self.lh.return_tips96()


class TestLiquidHandlerSerializeState(unittest.IsolatedAsyncioTestCase):
  """Tests for LiquidHandler.serialize_state() and load_state()."""

  async def asyncSetUp(self):
    self.backend = _create_mock_backend(num_channels=8)
    self.deck = STARLetDeck()
    self.lh = LiquidHandler(backend=self.backend, deck=self.deck)
    self.tip_rack = hamilton_96_tiprack_300uL_filter(name="tip_rack")
    self.plate = Cor_96_wellplate_360ul_Fb(name="plate")
    self.deck.assign_child_resource(self.tip_rack, location=Coordinate(0, 0, 0))
    self.deck.assign_child_resource(self.plate, location=Coordinate(100, 100, 0))
    await self.lh.setup()

  async def test_serialize_state_after_setup(self):
    state = self.lh.serialize_state()
    self.assertIn("head_state", state)
    self.assertIn("head96_state", state)
    self.assertIn("arm_state", state)

    # 8 channels, all without tips
    self.assertEqual(len(state["head_state"]), 8)
    for channel_state in state["head_state"].values():
      self.assertIsNone(channel_state["tip"])

    # 96 channels for head96
    self.assertEqual(len(state["head96_state"]), 96)

    # 1 arm, no resource picked up
    self.assertEqual(state["arm_state"], {0: None})

  async def test_serialize_state_no_head96(self):
    backend = _create_mock_backend(num_channels=8)
    type(backend).head96_installed = PropertyMock(return_value=False)
    deck = STARLetDeck()
    lh = LiquidHandler(backend=backend, deck=deck)
    await lh.setup()

    state = lh.serialize_state()
    self.assertIsNone(state["head96_state"])

  async def test_serialize_state_no_arms(self):
    backend = _create_mock_backend(num_channels=8)
    type(backend).num_arms = PropertyMock(return_value=0)
    deck = STARLetDeck()
    lh = LiquidHandler(backend=backend, deck=deck)
    await lh.setup()

    state = lh.serialize_state()
    self.assertIsNone(state["arm_state"])

  async def test_serialize_state_with_resource_pickup(self):
    resource = self.plate
    pickup = ResourcePickup(
      resource=resource,
      offset=Coordinate.zero(),
      pickup_distance_from_top=5.0,
      direction=GripDirection.FRONT,
    )
    self.lh._resource_pickup = pickup

    state = self.lh.serialize_state()
    arm_state = state["arm_state"]
    self.assertIsNotNone(arm_state[0])

    serialized_pickup = arm_state[0]
    self.assertEqual(serialized_pickup["type"], "ResourcePickup")
    self.assertEqual(serialized_pickup["pickup_distance_from_top"], 5.0)
    self.assertEqual(serialized_pickup["direction"], "FRONT")
    self.assertIn("resource", serialized_pickup)
    self.assertEqual(serialized_pickup["resource"]["name"], "plate")

  async def test_serialize_state_arm_none_after_drop(self):
    self.lh._resource_pickup = ResourcePickup(
      resource=self.plate,
      offset=Coordinate.zero(),
      pickup_distance_from_top=5.0,
      direction=GripDirection.LEFT,
    )
    self.assertIsNotNone(self.lh.serialize_state()["arm_state"][0])

    self.lh._resource_pickup = None
    self.assertEqual(self.lh.serialize_state()["arm_state"], {0: None})

  async def test_load_state_head(self):
    state = self.lh.serialize_state()
    # Modify a tracker, load old state, verify it's restored
    self.lh.head[0]._tip = hamilton_96_tiprack_300uL_filter(name="tmp").get_item("A1").get_tip()
    self.lh.load_state(state)
    self.assertIsNone(self.lh.head[0]._tip)

  async def test_load_state_backward_compatible(self):
    # Old state format without head96_state or arm_state
    old_state = {"head_state": {c: self.lh.head[c].serialize() for c in self.lh.head}}
    self.lh.load_state(old_state)  # should not raise



from pylabrobot.resources.height_volume_functions import (
  calculate_liquid_volume_container_2segments_square_vbottom,
)
from pylabrobot.resources.plate import Lid, Plate
from pylabrobot.resources.utils import create_ordered_items_2d
from pylabrobot.resources.well import (
  CrossSectionType,
  Well,
  WellBottomType,
)

# # # # # # # # # # Cor_Axy_24_wellplate_10mL_Vb # # # # # # # # # #


def Cor_Axy_24_wellplate_10mL_Vb(name: str, with_lid: bool = False) -> Plate:
  """
  Corning cat. no.: P-DW-10ML-24-C-S
  - manufacturer_link: https://ecatalog.corning.com/life-sciences/b2b/UK/en/Genomics-&-Molecular-Biology/Automation-Consumables/Deep-Well-Plate/Axygen%C2%AE-Deep-Well-and-Assay-Plates/p/P-DW-10ML-24-C
  - brand: Axygen
  - distributor: (Fisher Scientific, 12557837)
  - material: Polypropylene
  - sterile: yes
  - autoclavable: yes
  """
  return Plate(
    name=name,
    size_x=127.76,
    size_y=85.48,
    size_z=44.24,
    lid=Cor_Axy_24_wellplate_10mL_Vb_Lid(name + "_lid") if with_lid else None,
    model=Cor_Axy_24_wellplate_10mL_Vb.__name__,
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=6,
      num_items_y=4,
      dx=9.8,
      dy=7.2,
      dz=1.2,
      item_dx=18,
      item_dy=18,
      size_x=17.0,
      size_y=17.0,
      size_z=42,
      material_z_thickness=1.46,
      bottom_type=WellBottomType.V,
      compute_volume_from_height=_compute_volume_from_height_Cor_Axy_24_wellplate_10mL_Vb,
      cross_section_type=CrossSectionType.RECTANGLE,
    ),
  )


def Cor_Axy_24_wellplate_10mL_Vb_Lid(name: str) -> Lid:
  raise NotImplementedError("This lid is not currently defined.")


def _compute_volume_from_height_Cor_Axy_24_wellplate_10mL_Vb(h: float):
  if h > 42.1:
    raise ValueError(f"Height {h} is too large for Cos_96_Vb")
  return calculate_liquid_volume_container_2segments_square_vbottom(
    x=17, y=17, h_pyramid=5, h_cube=37, liquid_height=h
  )


# # # # # # # # # # Cor_Axy_96_wellplate_500uL_Ub # # # # # # # # # #


def Cor_Axy_96_wellplate_500uL_Ub(name: str, with_lid: bool = False) -> Plate:
  """
  Axygen 96w Shallow Well Plate 500uL U Bottom
  - Product number: P-96-450V-C-S
  - manufacturer_link: https://ecatalog.corning.com/life-sciences/b2c/US/en/Genomics-&-Molecular-Biology/Automation-Consumables/Deep-Well-Plate/Axygen%C2%AE-Deep-Well-and-Assay-Plates/p/P-96-450V-C-S
  - Spec sheet info: https://www.corning.com/catalog/cls/documents/selection-guides/CLS-A-PSG-001.pdf
  - working volume: 450uL
  - brand: Axygen
  - material: Polypropylene
  - sterile: yes
  """
  return Plate(
    name=name,
    size_x=127.0,  # measured
    size_y=85.51,  # measured
    size_z=14.82,  # measured
    model="Cor_Axy_96_wellplate_500uL_Ub",
    ordered_items=create_ordered_items_2d(
      Well,
      num_items_x=12,  # from spec
      num_items_y=8,  # from spec
      dx=9.6,  # measured
      dy=7.0,  # measured
      dz=1.2,  # measured
      item_dx=9.0,  # measured
      item_dy=9.0,  # measured
      size_x=8.0,  # measured
      size_y=8.0,  # measured
      size_z=14.82 - 2.57,  # measured
      bottom_type=WellBottomType.U,
      material_z_thickness=1.18,  # measured
      cross_section_type=CrossSectionType.CIRCLE,
    ),
  )



from pylabrobot.resources.carrier import (
  ResourceHolder,
  TubeCarrier,
  create_resources,
)
from pylabrobot.resources.coordinate import Coordinate


def boekel_50mL_falcon_carrier(name: str) -> TubeCarrier:
  """50 mL Falcon tube carrier orientation.

  https://www.boekelsci.com/multi-tube-rack-for-50ml-conical-15ml-conical-and-
  microcentrifuge-tubes-pn-120008.html.

  Oriented in landscape mode with the male connector to the right.

  Sizes from website; carrier site locations measured with ruler.
  """

  return TubeCarrier(
    name=name,
    size_x=174,
    size_y=52,
    size_z=95,
    sites=create_resources(
      klass=ResourceHolder,
      locations=[Coordinate(x=x, y=11, z=5) for x in [11, 46, 91, 127]],
      resource_size_x=[30] * 4,
      resource_size_y=[30] * 4,
    ),
    model="Boekel Scientific Tube Carrier",
  )


def boekel_15mL_falcon_carrier(name: str) -> TubeCarrier:
  """15 mL Falcon tube carrier orientation.

  https://www.boekelsci.com/multi-tube-rack-for-50ml-conical-15ml-conical-and-
  microcentrifuge-tubes-pn-120008.html

  Oriented in landscape mode with the male connector to the right.

  Sizes from website; carrier site locations measured with ruler.
  """

  return TubeCarrier(
    name=name,
    size_x=174,
    size_y=52,
    size_z=95,
    sites=create_resources(
      klass=ResourceHolder,
      locations=[Coordinate(x=x, y=27, z=5) for x in [5, 34, 63, 88, 118, 147]]
      + [Coordinate(x=x, y=4.5, z=5) for x in [5, 34, 63, 88, 118, 147]],
      resource_size_x=[17] * 16,
      resource_size_y=[17] * 16,
    ),
    model="Boekel Scientific Tube Carrier",
  )


def boekel_1_5mL_microcentrifuge_carrier(name: str) -> TubeCarrier:
  """1.5 mL microcentrifuge tube carrier orientation.

  https://www.boekelsci.com/multi-tube-rack-for-50ml-conical-15ml-conical-and-
  microcentrifuge-tubes-pn-120008.html

  Oriented in landscape mode with the male connector to the right.

  Sizes from website; carrier site locations measured with ruler.
  """

  x_locs = [4, 25, 46, 67, 88, 109, 131, 152]

  return TubeCarrier(
    name=name,
    size_x=174,
    size_y=52,
    size_z=95,
    sites=create_resources(
      klass=ResourceHolder,
      locations=[Coordinate(x=x, y=57, z=5) for x in x_locs]
      + [Coordinate(x=x, y=48, z=5) for x in x_locs]
      + [Coordinate(x=x, y=39, z=5) for x in x_locs]
      + [Coordinate(x=x, y=10, z=5) for x in x_locs],
      resource_size_x=[13] * 32,
      resource_size_y=[13] * 32,
    ),
    model="Boekel Scientific Tube Carrier",
  )


def boekel_mini_microcentrifuge_carrier(name: str) -> TubeCarrier:
  """The tiniest microcentrifuge tube carrier orientation.

  https://www.boekelsci.com/multi-tube-rack-for-50ml-conical-15ml-conical-and-
  microcentrifuge-tubes-pn-120008.html

  Oriented in landscape mode with the male connector to the right.

  Sizes from website; carrier site locations measured with ruler.
  """

  x_locs = [5, 27, 48, 70, 91, 113, 134, 154]

  return TubeCarrier(
    name=name,
    size_x=174,
    size_y=52,
    size_z=95,
    sites=create_resources(
      klass=ResourceHolder,
      locations=[Coordinate(x=x, y=68.5, z=5) for x in x_locs]
      + [Coordinate(x=x, y=50, z=5) for x in x_locs]
      + [Coordinate(x=x, y=31, z=5) for x in x_locs]
      + [Coordinate(x=x, y=12, z=5) for x in x_locs],
      resource_size_x=[9] * 32,
      resource_size_y=[9] * 32,
    ),
    model="Boekel Scientific Tube Carrier",
  )



import xml.etree.ElementTree as ET
from typing import Any, Dict, Literal, Optional

from pylabrobot.machines.backend import MachineBackend
from pylabrobot.storage.inheco.scila.inheco_sila_interface import InhecoSiLAInterface


def _parse_scalar(text: Optional[str], tag: str) -> object:
  if text is None:
    return None
  s = text.strip()
  t = tag.lower()
  if t in ("float64", "float32", "double", "single"):
    return float(s)
  if t in ("int32", "int64", "uint32", "uint64", "integer"):
    return int(s)
  if t in ("boolean", "bool"):
    return s.lower() == "true"
  return s  # String or unknown => raw text


def _get_param(root: ET.Element, name: str):
  p = root.find(f".//Parameter[@name='{name}']")
  if p is None or len(p) == 0:
    raise RuntimeError(f"Response missing parameter '{name}'")
  child = next(iter(p))  # e.g. <Float64>, <Boolean>, <String>
  return _parse_scalar(child.text, child.tag)


def _get_params(root: ET.Element, names: list[str]) -> dict[str, object]:
  return {n: _get_param(root, n) for n in names}


DrawerStatus = Literal["Opened", "Closed"]


class SCILABackend(MachineBackend):
  def __init__(self, scila_ip: str, client_ip: Optional[str] = None) -> None:
    self._sila_interface = InhecoSiLAInterface(client_ip=client_ip, machine_ip=scila_ip)

  async def setup(self) -> None:
    await self._sila_interface.setup()
    await self._reset_and_initialize()

  async def stop(self) -> None:
    await self._sila_interface.close()

  async def _reset_and_initialize(self) -> None:
    event_uri = f"http://{self._sila_interface.client_ip}:{self._sila_interface.bound_port}/"
    await self._sila_interface.send_command(
      command="Reset", deviceId="MyController", eventReceiverURI=event_uri, simulationMode=False
    )

    await self._sila_interface.send_command("Initialize")

  async def request_status(self) -> str:
    # GetStatus returns synchronously (return_code 1 = immediate dict), unlike other commands
    # which return asynchronously (return_code 2 = XML via callback).
    resp = await self._sila_interface.send_command("GetStatus")
    return resp.get("GetStatusResponse", {}).get("state", "Unknown")  # type: ignore

  async def request_liquid_level(self) -> str:
    root = await self._sila_interface.send_command("GetLiquidLevel")
    return _get_param(root, "LiquidLevel")  # type: ignore

  async def request_temperature_information(self) -> dict[str, Any]:
    root = await self._sila_interface.send_command("GetTemperature")
    return _get_params(root, ["CurrentTemperature", "TargetTemperature", "TemperatureControl"])  # type: ignore

  async def measure_temperature(self) -> float:
    return (await self.request_temperature_information())["CurrentTemperature"]  # type: ignore

  async def request_target_temperature(self) -> float:
    return (await self.request_temperature_information())["TargetTemperature"]  # type: ignore

  async def is_temperature_control_enabled(self) -> bool:
    return (await self.request_temperature_information())["TemperatureControl"]  # type: ignore

  async def open(self, drawer_id: int) -> None:
    if drawer_id not in {1, 2, 3, 4}:
      raise ValueError(f"Invalid drawer ID: {drawer_id}. Must be 1, 2, 3, or 4.")
    await self._sila_interface.send_command("PrepareForInput", position=drawer_id)
    await self._sila_interface.send_command("OpenDoor")

  async def close(self, drawer_id: int) -> None:
    if drawer_id not in {1, 2, 3, 4}:
      raise ValueError(f"Invalid drawer ID: {drawer_id}. Must be 1, 2, 3, or 4.")
    await self._sila_interface.send_command("PrepareForOutput", position=drawer_id)
    await self._sila_interface.send_command("CloseDoor")

  async def request_drawer_statuses(self) -> Dict[int, DrawerStatus]:
    root = await self._sila_interface.send_command("GetDoorStatus")
    params = _get_params(root, ["Drawer1", "Drawer2", "Drawer3", "Drawer4"])
    return {i: params[f"Drawer{i}"] for i in range(1, 5)}  # type: ignore

  async def request_drawer_status(self, drawer_id: int) -> DrawerStatus:
    if drawer_id not in {1, 2, 3, 4}:
      raise ValueError(f"Invalid drawer ID: {drawer_id}. Must be 1, 2, 3, or 4.")
    positions = await self.request_drawer_statuses()
    return positions[drawer_id]

  async def request_co2_flow_status(self) -> str:
    root = await self._sila_interface.send_command("GetCO2FlowStatus")
    return _get_param(root, "CO2FlowStatus")  # type: ignore

  async def request_valve_status(self) -> dict[str, str]:
    """
    example:

    {
      "H2O": "Opened",
      "CO2 Normal": "Opened",
      "CO2 Boost": "Closed"
    }
    """

    root = await self._sila_interface.send_command("GetValveStatus")
    return _get_params(root, ["H2O", "CO2 Normal", "CO2 Boost"])  # type: ignore

  async def start_temperature_control(self, temperature: float) -> None:
    await self._sila_interface.send_command(
      "SetTemperature", targetTemperature=temperature, temperatureControl=True
    )

  async def stop_temperature_control(self) -> None:
    await self._sila_interface.send_command("SetTemperature", temperatureControl=False)

  def serialize(self) -> dict[str, Any]:
    return {
      **super().serialize(),
      "scila_ip": self._sila_interface.machine_ip,
      "client_ip": self._sila_interface.client_ip,
    }

  @classmethod
  def deserialize(cls, data: dict[str, Any]) -> "SCILABackend":
    return cls(scila_ip=data["scila_ip"], client_ip=data.get("client_ip"))



"""A simple JSON serializer."""

import enum
import inspect
import marshal
import math
import sys
import types
from typing import Any, Dict, List, Union, cast

if sys.version_info >= (3, 10):
  from typing import TypeAlias
else:
  from typing_extensions import TypeAlias

from pylabrobot.utils.object_parsing import find_subclass

JSON: TypeAlias = Union[Dict[str, "JSON"], List["JSON"], str, int, float, bool, None]


class SerializableMixin:
  """Mixin for classes that can be serialized and deserialized."""

  def serialize(self) -> dict:
    data = {}
    for key, value in vars(self).items():
      if key.startswith("_"):
        continue
      data[key] = serialize(value)
    data["type"] = self.__class__.__name__
    return data


def serialize(obj: Any) -> JSON:
  """Serialize an object."""

  if isinstance(obj, (int, float, str, bool, type(None))):
    # infinities and NaNs are not valid JSON, so we convert them to strings
    if isinstance(obj, float) and not math.isfinite(obj):
      return "nan" if math.isnan(obj) else ("Infinity" if obj > 0 else "-Infinity")
    return obj
  if isinstance(obj, (list, tuple, set)):
    return [serialize(item) for item in obj]
  if isinstance(obj, dict):
    return {k: serialize(v) for k, v in obj.items()}
  if isinstance(obj, enum.Enum):
    return obj.name
  if inspect.isfunction(obj):
    return {
      "type": "function",
      "code": marshal.dumps(obj.__code__).hex(),
      "closure": serialize(obj.__closure__) if obj.__closure__ else None,
    }
  if isinstance(obj, types.CellType):
    return {"type": "cell", "contents": serialize(obj.cell_contents)}
  if isinstance(obj, object):
    if hasattr(obj, "serialize"):  # if the object has a custom serialize method
      return cast(JSON, obj.serialize())
    else:
      data: Dict[str, Any] = {}
      for key, value in vars(obj).items():
        if key.startswith("_"):
          continue
        data[key] = serialize(value)
      data["type"] = obj.__class__.__name__
      return data
  raise TypeError(f"Cannot serialize {obj} of type {type(obj)}")


def deserialize(data: JSON, allow_marshal: bool = False) -> Any:
  """Deserialize an object."""

  if isinstance(data, str):
    if data == "Infinity":
      return math.inf
    if data == "-Infinity":
      return -math.inf
    if data == "nan":
      return math.nan
    return data
  if isinstance(data, (int, float, bool, type(None))):
    return data
  if isinstance(data, list):
    return [deserialize(item, allow_marshal=allow_marshal) for item in data]
  if isinstance(data, dict):
    if "type" in data:  # deserialize a class
      data = data.copy()
      klass_type = cast(str, data.pop("type"))
      if klass_type == "function":
        if allow_marshal:
          assert isinstance(data["code"], str)
          code = marshal.loads(bytes.fromhex(data["code"]))
          closure = (
            tuple(deserialize(data["closure"], allow_marshal=allow_marshal))
            if data["closure"]
            else None
          )
          return types.FunctionType(code, globals(), closure=closure)
        return None
      if klass_type == "cell":
        return types.CellType(deserialize(data["contents"], allow_marshal=allow_marshal))
      klass = find_subclass(klass_type, cls=SerializableMixin)
      if klass is None:
        raise ValueError(f"Could not find class '{klass_type}'")
      params = {k: deserialize(v, allow_marshal=allow_marshal) for k, v in data.items()}
      if "deserialize" in klass.__dict__:
        return klass.deserialize(params)  # type: ignore[attr-defined]
      return klass(**params)
    return {k: deserialize(v, allow_marshal=allow_marshal) for k, v in data.items()}
  if isinstance(data, object):
    return data
  raise TypeError(f"Cannot deserialize {data} of type {type(data)}")



import unittest
import unittest.mock

from pylabrobot.machines.machine import Machine, MachineBackend


class TestMachine(unittest.TestCase):
  class MockBackend(MachineBackend):
    def __init__(self, mock_param):
      self.mock_param = mock_param

    async def setup(self):
      pass

    async def stop(self):
      pass

    def serialize(self):
      return {**super().serialize(), "mock_param": self.mock_param}

  class MockMachine(Machine):
    pass

  def test_serialize(self):
    m = self.MockMachine(backend=self.MockBackend("mock_param"))
    self.assertEqual(
      m.serialize(),
      {
        "backend": {
          "mock_param": "mock_param",
          "type": "MockBackend",
        },
      },
    )

  def test_deserialize(self):
    m = self.MockMachine(backend=self.MockBackend("mock_param"))
    Machine.deserialize(m.serialize())  # shouldn't raise



class LoaderNoPlateError(Exception):
  pass


class CentrifugeDoorError(Exception):
  pass


class NotAtBucketError(Exception):
  pass


class BucketNoPlateError(Exception):
  pass


class BucketHasPlateError(Exception):
  pass
